"""
canvas
===============================================================================

Implementation of the Canvas in a Glyph View.
"""
from __future__ import annotations

import logging
import os
import weakref
from typing import TYPE_CHECKING, Dict, Optional

import wx
from fontTools.misc.transform import Transform
from wbBase.document import dbg
from wbDefcon import Color, Font, Glyph, Guideline

from ...dialog.findGlyphDialog import FindGlyphDialog
from .drawing import FontLevelPlanes, LayerPlanes
from .tool import (
    AddEllipse,
    AddRectangle,
    MagicWand,
    ToolDraw,
    ToolEdit,
    ToolErase,
    ToolKnife,
    ToolMeter,
    ToolPan,
    ToolReverse,
    ToolRotate,
    ToolScale,
    ToolStartpoint,
    ToolZoom,
)
from .tool.base import WORKING

if TYPE_CHECKING:
    from .editPanel import GlyphEditPanel
    from .tool.base import GlyphTool

log = logging.getLogger(__name__)

DISPLAY_NORMAL = 0
DISPLAY_PREVIEW = 1


class Canvas(wx.ScrolledWindow):
    """
    The drawing area of the glyph editor
    """

    refreshNotifications = (
        "Glyph.Changed",
        # "Glyph.NameChanged",
        "Glyph.WidthChanged",
        "Glyph.ContoursChanged",
        "Glyph.AnchorsChanged",
        "Glyph.GuidelinesChanged",
        "Glyph.ComponentsChanged",
        "Glyph.ComponentBaseGlyphDataChanged",
    )
    fontNotifications = ("Font.GuidelinesChanged",)
    layerSetNotifications = (
        "LayerSet.LayerOrderChanged",
        "LayerSet.LayerAdded",
        "LayerSet.LayerDeleted",
        "LayerSet.LayerNameChanged",
    )

    def __init__(self, parent, doc=None):
        style = 0
        wx.ScrolledWindow.__init__(
            self,
            parent,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.DefaultSize,
            style,
            "GlyphEditCanvas",
        )
        self.ClientSize: wx.Size
        self.Parent: GlyphEditPanel
        self.doc = doc
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self.SetBackgroundColour("WHITE")
        self.canvasSize = wx.Size(6000, 10000)  # in font units
        self.canvasOrigin = wx.Point(2000, 4000)  # in font units
        self.SetScrollRate(1, 1)
        self._zoom = 0.5  # screen pixel per font unit
        self._minZoom = 0.05
        self._maxZoom = 20
        self._font: Optional[Font] = None
        self._glyph: Optional[Glyph] = None
        self.transform = None
        self.drawingPlanes = []
        self.tools: Dict[str, GlyphTool] = {}
        self._tool = None
        self._prevTool = None
        self.glyphCommands = ["FlipHorizontal", "FlipVertical"]
        self.displayMode = DISPLAY_NORMAL

        # Connect Events
        # window events
        self.Bind(wx.EVT_ENTER_WINDOW, self.on_ENTER_WINDOW)
        # wx.EVT_LEAVE_WINDOW(self, self.on_LEAVE_WINDOW)
        self.Bind(wx.EVT_SET_FOCUS, self.on_SET_FOCUS)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.on_SIZE)
        self.Bind(wx.EVT_SCROLLWIN, self.on_SCROLLWIN)
        self.Bind(wx.EVT_MOUSE_CAPTURE_LOST, self.on_MOUSE_CAPTURE_LOST)

        # mouse events
        self.Bind(wx.EVT_MOUSEWHEEL, self.on_MOUSEWHEEL)
        # wx.EVT_RIGHT_DOWN(self, self.on_RIGHT_DOWN)

        # keyboard events
        self.Bind(wx.EVT_KEY_DOWN, self.on_KEY_DOWN)
        self.Bind(wx.EVT_KEY_UP, self.on_KEY_UP)

        self._initTools()

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def Destroy(self):
        del self.glyph
        wx.LogDebug("Canvas.Destroy()")
        return super(Canvas, self).Destroy()

    # -------------------------------------------------------------------------
    # properties
    # -------------------------------------------------------------------------

    @property
    def app(self):
        return wx.GetApp()

    @property
    def view(self):
        return self.Parent.Parent.view

    @property
    def font(self) -> Optional[Font]:
        if self._font is None:
            return None
        return self._font()

    @font.setter
    def font(self, font):
        if font is None:
            del self.font
            return
        assert isinstance(font, Font)
        currentFont = self.font
        if font != currentFont:
            del self.font
            for notificationName in self.fontNotifications:
                font.addObserver(
                    self, "handleNotification", notification=notificationName
                )
            layerSet = font.layers
            for notificationName in self.layerSetNotifications:
                layerSet.addObserver(
                    self, "handleNotification", notification=notificationName
                )
            self._font = weakref.ref(font)
        self.updateDrawingPlanes()

    @font.deleter
    def font(self):
        currentFont = self.font
        if isinstance(currentFont, Font):
            for notificationName in self.fontNotifications:
                currentFont.removeObserver(self, notificationName)
            for notificationName in self.layerSetNotifications:
                currentFont.layers.removeObserver(self, notificationName)
        self._font = None

    @property
    def glyph(self) -> Optional[Glyph]:
        if self._glyph is None:
            return None
        return self._glyph()

    @glyph.setter
    def glyph(self, glyph: Optional[Glyph]):
        dbg(f"Canvas set glyph {glyph}")
        if glyph is None:
            del self.glyph
            return
        assert isinstance(glyph, Glyph)
        assert glyph.font == self.font
        if self.glyph == glyph:
            return
        del self.glyph
        for layer in glyph.layerSet:
            for notificationName in self.refreshNotifications:
                if glyph.name in layer and not layer[glyph.name].hasObserver(
                    self, notificationName
                ):
                    layer[glyph.name].addObserver(
                        self, "handleNotification", notificationName
                    )
        self._glyph = weakref.ref(glyph)
        self.Refresh()

    @glyph.deleter
    def glyph(self):
        currentGlyph = self.glyph
        if isinstance(currentGlyph, Glyph):
            for layer in currentGlyph.layerSet:
                for notificationName in self.refreshNotifications:
                    if currentGlyph.name in layer and layer[
                        currentGlyph.name
                    ].hasObserver(self, notificationName):
                        layer[currentGlyph.name].removeObserver(self, notificationName)
        self._glyph = None

    def handleNotification(self, notification):
        log.debug("Canvas.handleNotification %r", notification)
        if notification.name.startswith("LayerSet."):
            self.updateDrawingPlanes()
        try:
            self.Refresh()
        except RuntimeError:
            wx.LogDebug("Refresh failed in Canvas.handleNotification")

    def showNextGlyph(self) -> None:
        """Show next glyph based on glyph order in glyphGridPanel."""
        glyphNames = self.font.document.frame.glyphGridPanel.glyphNames
        currentIndex = glyphNames.index(self.glyph.name)
        if currentIndex < len(self.font) - 1:
            self.view.glyph = self.font[glyphNames[currentIndex + 1]]

    def showPreviousGlyph(self) -> None:
        """Show previous glyph based on glyph order in glyphGridPanel."""
        glyphNames = self.font.document.frame.glyphGridPanel.glyphNames
        currentIndex = glyphNames.index(self.glyph.name)
        if currentIndex > 0:
            self.view.glyph = self.font[glyphNames[currentIndex - 1]]

    # Drawing Planes ----------------------------

    @property
    def layerPlanes(self) -> LayerPlanes:
        return tuple(p for p in self.drawingPlanes if isinstance(p, LayerPlanes))

    @property
    def fontLevelPlanesStack(self):
        for planeStack in self.drawingPlanes:
            if isinstance(planeStack, FontLevelPlanes):
                return planeStack

    def updateDrawingPlanes(self):
        oldPlanes = {}
        newPlanes = []
        for planeStack in self.drawingPlanes:
            if isinstance(planeStack, FontLevelPlanes):
                newPlanes.append(planeStack)
            elif isinstance(planeStack, LayerPlanes):
                oldPlanes[planeStack.name] = planeStack
        if not newPlanes:
            newPlanes.append(FontLevelPlanes(self))
        for layerName in self.font.layers.layerOrder:
            if layerName in oldPlanes:
                newPlanes.append(oldPlanes[layerName])
            else:
                newPlanes.append(LayerPlanes(self, layerName))
        self.drawingPlanes = newPlanes
        self.getActiveDrawingPlaneStack()

    def getDrawingPlaneStack(self, name):
        for planeStack in self.drawingPlanes:
            if planeStack.name == name:
                return planeStack

    def getActiveDrawingPlaneStack(self):
        for planeStack in self.drawingPlanes:
            if planeStack.active:
                return planeStack
        planeStack = self.getDrawingPlaneStack(self.font.layers.defaultLayer.name)
        planeStack.active = True
        return planeStack

    # UnDo / ReDo handling ----------------------------

    def CanUndo(self):
        return self.glyph.canUndo()

    def Undo(self):
        self.glyph.undo()

    def CanRedo(self):
        return self.glyph.canRedo()

    def Redo(self):
        self.glyph.redo()

    # Find ---------------------------------------

    def CanFind(self):
        return True

    def doFind(self):
        # wx.LogInfo('Ctrl + F pressed')
        with FindGlyphDialog(self, self.glyph.layer) as findGlyphDialog:
            findGlyphDialog.layer = self.glyph.layer
            if findGlyphDialog.ShowModal() == wx.ID_OK:
                if findGlyphDialog.selctedGlyph is not None:
                    self.font[findGlyphDialog.selctedGlyph].show(view=self.view)
                elif (
                    findGlyphDialog.checkBox_create.Value
                    and findGlyphDialog.choice_findAttr.StringSelection == "Name"
                ):
                    newGlyphName = findGlyphDialog.textCtrl_findValue.Value
                    if newGlyphName not in self.font:
                        newGlyph = self.font.newGlyph(newGlyphName)
                        newGlyph.show(view=self.view)

    # Zooming ---------------------------------------

    @property
    def zoom(self):
        return self._zoom

    @zoom.setter
    def zoom(self, value):
        newVal = min(max(round(float(value), 2), self._minZoom), self._maxZoom)
        if newVal != self._zoom:
            self._zoom = newVal
            self.SetVirtualSize(
                wx.Size(round(self.canvasSize.width * newVal), round(self.canvasSize.height * newVal))
            )

    @zoom.deleter
    def zoom(self):
        self.zoom = 1

    def CanZoomIn(self):
        return self.zoom < self._maxZoom

    def ZoomIn(self):
        self.Freeze()
        x = self.screenToCanvasX(self.ClientSize.width / 2)
        y = self.screenToCanvasY(self.ClientSize.height / 2)
        self.zoom *= 1.2
        self.centerCanvasOnScreen(x, y)
        self.Thaw()

    def CanZoomOut(self):
        return self.zoom > self._minZoom

    def ZoomOut(self):
        self.Freeze()
        x = self.screenToCanvasX(self.ClientSize.width / 2)
        y = self.screenToCanvasY(self.ClientSize.height / 2)
        self.zoom *= 0.8
        self.centerCanvasOnScreen(x, y)
        self.Thaw()

    def CanZoom100(self):
        return self.zoom != 1.0

    def Zoom100(self):
        self.Freeze()
        x = self.screenToCanvasX(self.ClientSize.width / 2)
        y = self.screenToCanvasY(self.ClientSize.height / 2)
        self.zoom = 1.0
        self.centerCanvasOnScreen(x, y)
        self.Thaw()

    # tool handling -----------------------------------------------------------

    def _initTools(self):
        self.tools["ToolEdit"] = ToolEdit(self)
        self.tools["ToolDraw"] = ToolDraw(self)
        self.tools["ToolZoom"] = ToolZoom(self)
        self.tools["ToolPan"] = ToolPan(self)
        self.tools["MagicWand"] = MagicWand(self)
        self.tools["AddRectangle"] = AddRectangle(self)
        self.tools["AddEllipse"] = AddEllipse(self)
        self.tools["ToolRotate"] = ToolRotate(self)
        self.tools["ToolScale"] = ToolScale(self)
        self.tools["ToolErase"] = ToolErase(self)
        self.tools["ToolMeter"] = ToolMeter(self)
        self.tools["ToolKnife"] = ToolKnife(self)
        self.tools["ToolStartpoint"] = ToolStartpoint(self)
        self.tools["ToolReverse"] = ToolReverse(self)
        self._tool = "ToolEdit"
        self._prevTool = "ToolEdit"
        self.tools["ToolEdit"].activate()

    @property
    def currentTool(self) -> GlyphTool:
        return self.tools[self._tool]

    def has_Tool(self, name: str) -> bool:
        return name in self.tools

    def getTool(self, name):
        return self.tools.get(name)

    def isToolActive(self, name):
        """
        Is the tool with the given name active?
        """
        return self._tool == name

    def selectTool(self, name):
        """
        Select the tool given by name.
        """
        if name in self.tools and name != self._tool:
            self.tools[self._tool].activate(False)
            self._prevTool = self._tool
            self._tool = name
            self.tools[self._tool].activate(True)

    def unselectTool(self):
        """
        Unselect the current tool and reselect the previus one.
        """
        self.tools[self._tool].activate(False)
        self._tool = self._prevTool
        self.tools[self._tool].activate(True)
        self._prevTool = "ToolEdit"

    # # command handling -----------------------------------------------------------

    # def has_GlyphCommand(self, commandName):
    #     return commandName in self.glyphCommands

    # def applyGlyphCommand(self, commandName):
    #     if commandName in self.glyphCommands:
    #         wx.LogWarning('Glyph Command "%s" not yet implemented' % commandName)

    # calculate canvas coordinates to screen coordinates ------------------------

    @property
    def ViewStart(self):
        return self.GetViewStart()

    def canvasToScreenX(self, value):
        return int(round((value + self.canvasOrigin.x) * self.zoom - self.ViewStart[0]))

    def canvasToScreenXrel(self, value):
        return int(round(value * self.zoom))

    def canvasToScreenY(self, value):
        return int(
            round((value - self.canvasOrigin.y) * -self.zoom - self.ViewStart[1])
        )

    def canvasToScreenYrel(self, value):
        return int(round(value * -self.zoom))

    # calculate screen coordinates to canvas coordinates ------------------------

    def screenToCanvasX(self, value):
        return (value + self.ViewStart[0]) / self.zoom - self.canvasOrigin.x

    def screenToCanvasXrel(self, value):
        return value / self.zoom

    def screenToCanvasY(self, value):
        return self.canvasOrigin.y + (value + self.ViewStart[1]) / -self.zoom

    def screenToCanvasYrel(self, value):
        return value / -self.zoom

    # scroll the canvas ---------------------------------------------------------

    def moveCanvasOnScreen(self, canvas_x, canvas_y, screen_x, screen_y):
        """
        Try to scroll canvas position (canvas_x, canvas_y) [drawing units]
        to screen position (screen_x, screen_y) [Screen pixes]
        """
        dx = self.screenToCanvasX(screen_x) - canvas_x
        dy = self.screenToCanvasY(screen_y) - canvas_y
        self.Scroll(
            self.ViewStart[0] - self.canvasToScreenXrel(dx),
            self.ViewStart[1] - self.canvasToScreenYrel(dy),
        )
        self.Refresh()
        self.updateRuler()

    def centerCanvasOnScreen(self, canvas_x, canvas_y):
        self.moveCanvasOnScreen(
            canvas_x, canvas_y, self.ClientSize.x / 2, self.ClientSize.y / 2
        )

    def updateRuler(self):
        self.Parent.rulerTop.Refresh()
        self.Parent.rulerLeft.Refresh()

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_ENTER_WINDOW(self, event):
        if self.glyph is not None and isinstance(self.currentTool, ToolEdit):
            planes = self.getActiveDrawingPlaneStack()
            if planes["GlyphGuide"].locked:
                event.Skip()
                return
            guidelineParent = None
            if self.Parent.rulerTop.createGuideline:
                angle = 0
                guidelineParent = self.Parent.rulerTop.createGuideline
                self.Parent.rulerTop.createGuideline = None
            elif self.Parent.rulerLeft.createGuideline:
                angle = 90
                guidelineParent = self.Parent.rulerLeft.createGuideline
                self.Parent.rulerLeft.createGuideline = None
            else:
                event.Skip()
                return
            guideline = Guideline(
                guidelineDict=dict(
                    x=round(self.screenToCanvasX(event.X)),
                    y=round(self.screenToCanvasY(event.Y)),
                    angle=angle,
                )
            )
            if guidelineParent == "Glyph":
                glyph = self.glyph
                glyph.undoManager.saveState()
                glyph.disableNotifications()
                glyph.appendGuideline(guideline)
            elif guidelineParent == "Font":
                self.font.appendGuideline(guideline)
            planes["GlyphGuide"].visibleActive = True
            self.fontLevelPlanesStack["FontGuide"].visible = True
            tool = self.currentTool
            tool.unselectAll()
            tool.state = WORKING
            tool.subject = guideline
            if not self.HasCapture():
                self.CaptureMouse()
            self.SetFocus()
        event.Skip()

    def on_SET_FOCUS(self, event):
        if self.app.documentManager.currentView != self.view:
            wx.LogDebug(
                f"{self} on_SET_FOCUS: {self.app.documentManager.currentView} {self.view}"
            )
            self.view.Activate()
        event.Skip()

    def on_SIZE(self, event):
        self.Refresh()
        event.Skip()

    def on_SCROLLWIN(self, event):
        self.Refresh()
        self.updateRuler()
        event.Skip()

    def on_MOUSE_CAPTURE_LOST(self, event):
        self.ReleaseMouse()
        self.Refresh()
        event.Skip()

    def on_MOUSEWHEEL(self, event):
        if event.AltDown():
            val = event.GetWheelRotation()
            canvas_x = self.screenToCanvasX(event.GetX())
            canvas_y = self.screenToCanvasY(event.GetY())
            self.Freeze()
            if val > 0:
                self.zoom *= 1.2
            else:
                self.zoom *= 0.8
            self.moveCanvasOnScreen(canvas_x, canvas_y, event.GetX(), event.GetY())
            self.Thaw()
        event.Skip()

    def OnPaint(self, event):
        self.transform = (
            Transform()
            .translate(-self.ViewStart[0], -self.ViewStart[1])
            .scale(self.zoom, -self.zoom)
            .translate(self.canvasOrigin.x, -self.canvasOrigin.y)
        )
        bitmap = wx.Bitmap(self.ClientSize.width, self.ClientSize.height)
        dc = wx.BufferedPaintDC(self, bitmap)
        self.OnDraw(dc)

    def OnDraw(self, dc: wx.BufferedPaintDC):
        use_cairo = False
        activePlaneStack = self.getActiveDrawingPlaneStack()
        backgroundColour = wx.WHITE
        if activePlaneStack.layer.color:
            hue = activePlaneStack.layer.color.hue
            backgroundColour = Color.from_hsv(hue, 0.05, 1.0).wx
        dc.SetBackground(wx.TheBrushList.FindOrCreateBrush(backgroundColour))
        dc.Clear()
        if use_cairo:
            wxdir = os.path.dirname(wx.__file__) + os.pathsep
            if not wxdir in os.environ.get("PATH", ""):
                os.environ["PATH"] = wxdir + os.environ.get("PATH", "")
            gcr = wx.GraphicsRenderer.GetCairoRenderer()
            print(gcr)
            gc = gcr.CreateContext(dc)
        else:
            gc = wx.GraphicsContext.Create(dc)
        if self.displayMode == DISPLAY_NORMAL:
            # first: draw all visible and inactive PlaneStacks in backgroud
            for plane in [
                p for p in self.drawingPlanes if (p.visible and not p.active)
            ]:
                plane.draw(gc)
            # second: draw the active PlaneStack in foreground
            activePlaneStack.draw(gc)
            # third: let the current tool draw it's stuff
            self.currentTool.draw(gc)
        elif self.displayMode == DISPLAY_PREVIEW:
            glyphFill = activePlaneStack.get("GlyphFill")
            glyphFill.draw(gc)

    def on_KEY_DOWN(self, event):
        alt = event.AltDown()
        # cmd = event.CmdDown()
        ctrl = event.ControlDown()
        shift = event.ShiftDown()
        unicodeKey = event.GetUnicodeKey()
        if unicodeKey == wx.WXK_NONE:
            key = event.KeyCode
        else:
            key = chr(unicodeKey)
        if key == " ":  # Space pressed
            if not alt | ctrl | shift:
                self.selectTool("ToolPan")
            elif ctrl and not alt | shift:
                self.selectTool("ToolZoom")
        elif key == "Q":
            self.selectTool("ToolErase")
        elif key == "W":
            self.selectTool("ToolKnife")
        elif key == "T":
            self.selectTool("ToolMeter")
        elif key == "R":
            self.selectTool("AddRectangle")
        elif key == "E":
            self.selectTool("AddEllipse")
        elif key == "-" and not ctrl:
            self.Zoom100()
        elif key in ("+", wx.WXK_NUMPAD_ADD) and ctrl:
            self.ZoomIn()
        elif key in ("-", wx.WXK_NUMPAD_SUBTRACT) and ctrl:
            self.ZoomOut()
        elif key == "<" and self.displayMode == DISPLAY_NORMAL:
            self.displayMode = DISPLAY_PREVIEW
            self.Refresh()
        elif key == ",":
            self.showPreviousGlyph()
        elif key == ".":
            self.showNextGlyph()
        elif key == wx.WXK_CONTROL:
            if ctrl and not (alt | shift) and self.isToolActive("ToolPan"):
                self.selectTool("ToolZoom")
            elif ctrl and not alt and self.isToolActive("ToolEdit"):
                self.selectTool("MagicWand")
        elif key == wx.WXK_ALT:
            event.StopPropagation()
        if not alt:
            event.Skip()

    def on_KEY_UP(self, event):
        # alt = event.AltDown()
        # cmd = event.CmdDown()
        # ctrl = event.ControlDown()
        # shift = event.ShiftDown()
        unicodeKey = event.GetUnicodeKey()
        if unicodeKey == wx.WXK_NONE:
            key = event.KeyCode
        else:
            key = chr(unicodeKey)
        if key == " ":  # Space released
            if self.isToolActive("ToolPan"):
                self.unselectTool()
            if self.isToolActive("ToolZoom"):
                self.unselectTool()
        elif key in ("Q", "W", "E", "R", "T"):
            self.unselectTool()
        elif key == "<" and self.displayMode == DISPLAY_PREVIEW:
            self.displayMode = DISPLAY_NORMAL
            self.Refresh()
        elif key == wx.WXK_CONTROL:
            if self.isToolActive("ToolZoom"):
                self.unselectTool()
            if self.isToolActive("MagicWand"):
                self.unselectTool()
        elif key == wx.WXK_ALT:
            self.SetFocus()
            event.StopPropagation()
            return
        event.Skip()



"""
toolrotate
===============================================================================
"""

from .base import WORKING, TransformBaseTool


class ToolRotate(TransformBaseTool):
    def __init__(self, parent):
        TransformBaseTool.__init__(self, parent)
        self.anglePrev = 0
        self.angleSnap = 15

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------
    def on_LEFT_UP(self, event):
        self.anglePrev = 0
        TransformBaseTool.on_LEFT_UP(self, event)

    def on_MOTION(self, event):
        TransformBaseTool.on_MOTION(self, event)
        if event.Dragging() and self.state == WORKING:
            angle = self.angle
            delta = self.anglePrev - angle
            if delta:
                glyph = self.glyph
                if self.isAnySelected():
                    for contour in glyph:
                        for point in contour:
                            if point.selected:
                                point.rotateBy(delta, self.center)
                                self.contoursChanged = True
                    for anchor in glyph.anchors:
                        if anchor.selected:
                            anchor.rotateBy(delta, self.center)
                            self.anchorsChanged = True
                    for component in glyph.components:
                        if component.selected:
                            component.rotateBy(delta, self.center)
                            self.componentsChanged = True
                else:
                    # todo: check locked plains
                    glyph.rotateBy(delta, self.center)
                self.canvas.Refresh()
            self.anglePrev = angle
        event.Skip()



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version 3.10.0-4761b0c)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc

###########################################################################
## Class LayerDialogUI
###########################################################################

class LayerDialogUI ( wx.Dialog ):

    def __init__( self, parent ):
        wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = u"Layer", pos = wx.DefaultPosition, size = wx.DefaultSize, style = wx.CAPTION )

        self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )

        sizer = wx.FlexGridSizer( 0, 2, 0, 0 )
        sizer.AddGrowableCol( 1 )
        sizer.SetFlexibleDirection( wx.BOTH )
        sizer.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

        self.lbl_name = wx.StaticText( self, wx.ID_ANY, u"Name", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.lbl_name.Wrap( -1 )

        sizer.Add( self.lbl_name, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.textCtrl_name = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0 )
        self.textCtrl_name.SetMinSize( wx.Size( 150,-1 ) )

        sizer.Add( self.textCtrl_name, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.lbl_Color = wx.StaticText( self, wx.ID_ANY, u"Color", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.lbl_Color.Wrap( -1 )

        sizer.Add( self.lbl_Color, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

        self.colourPicker = wx.ColourPickerCtrl( self, wx.ID_ANY, wx.Colour( 255, 255, 255 ), wx.DefaultPosition, wx.DefaultSize, wx.CLRP_DEFAULT_STYLE )
        sizer.Add( self.colourPicker, 0, wx.ALL|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL, 5 )


        sizer.Add( ( 0, 0), 1, wx.EXPAND, 5 )

        sizer_button = wx.StdDialogButtonSizer()
        self.sizer_buttonOK = wx.Button( self, wx.ID_OK )
        sizer_button.AddButton( self.sizer_buttonOK )
        self.sizer_buttonCancel = wx.Button( self, wx.ID_CANCEL )
        sizer_button.AddButton( self.sizer_buttonCancel )
        sizer_button.Realize();

        sizer.Add( sizer_button, 1, wx.ALL|wx.EXPAND, 5 )


        self.SetSizer( sizer )
        self.Layout()
        sizer.Fit( self )

        self.Centre( wx.BOTH )

    def __del__( self ):
        pass





"""
base
===============================================================================
"""
import logging
from math import asin, hypot, pi, sqrt

import wx
from wbBase.tools import get_wxPen

from ..cursor import cursors
from ..drawing import LayerPlanes
from ..menu import DefaultEditMenu

log = logging.getLogger(__name__)

IDLE = 0
WORKING = 1
RUBBERBANDING = 2

DEGREE = 180 / pi


class GlyphTool(wx.EvtHandler):
    """
    Base class for glyph edit tools
    """

    cursor = cursors["Default"]
    pen = get_wxPen(wx.Colour(0, 0, 0, 255))
    brush = wx.NullBrush

    def __init__(self, parent):
        super().__init__()
        self.canvas = parent
        self.contoursChanged:bool = False
        self.anchorsChanged:bool = False
        self.componentsChanged:bool = False
        self.glyphGuidelinesChanged:bool = False
        self.fontGuidelinesChanged:bool = False
        self.state = IDLE
        self._active:bool = False
        self.subject = None
        self.x = 0
        self.y = 0
        self.xClick = None
        self.yClick = None
        self.xPrev = None
        self.yPrev = None
        self.shiftDown = False
        # Mouse Events
        self.Bind(wx.EVT_LEFT_DOWN, self.on_LEFT_DOWN)
        self.Bind(wx.EVT_LEFT_UP, self.on_LEFT_UP)
        self.Bind(wx.EVT_RIGHT_DOWN, self.on_RIGHT_DOWN)
        self.Bind(wx.EVT_MOTION, self.on_MOTION)

    def __repr__(self):
        return "<%s: active=%r>" % (self.__class__.__name__, self.active)

    @staticmethod
    def distance(p0, p1):
        dx = p1[0] - p0[0]
        dy = p1[1] - p0[1]
        return hypot(dx, dy)

    @property
    def activeStack(self):
        return self.canvas.getActiveDrawingPlaneStack()

    @property
    def glyph(self):
        stack = self.activeStack
        if isinstance(stack, LayerPlanes):
            return stack.glyph

    # Tool activation ------------------

    @property
    def active(self)->bool:
        """
        Active state of the tool
        """
        return self._active

    def activate(self, value:bool=True) -> None:
        assert isinstance(value, bool)
        if value:
            return self.onActivate()
        return self.onDeactivate()

    def onActivate(self):
        if not self._active:
            self.canvas.PushEventHandler(self)
            self.canvas.SetCursor(self.cursor)
            self.reset()
            self._active = True
            return True
        return False

    def onDeactivate(self):
        if self._active:
            self.canvas.PopEventHandler()
            self.canvas.SetCursor(cursors["Default"])
            self.reset()
            self._active = False
            return True
        return False

    def reset(self):
        self.x = 0
        self.y = 0
        self.xClick = None
        self.yClick = None
        self.xPrev = None
        self.yPrev = None
        self.state = IDLE
        self.shiftDown = False
        self.contoursChanged = False
        self.anchorsChanged = False
        self.componentsChanged = False
        self.glyphGuidelinesChanged = False
        self.fontGuidelinesChanged = False

    def hitTest(self):
        # check for pointing Stuff under cursor
        canvas_x = self.canvas.screenToCanvasX(self.x)
        canvas_y = self.canvas.screenToCanvasY(self.y)
        pointed = None
        self.subject = None
        # perform hit test on editable layers
        for plainName in self.activeStack.hitTestOrder:
            plain = self.activeStack.get(plainName)
            if plain.visible and not plain.locked:
                pointed = plain.hitTest(canvas_x, canvas_y)
                if pointed is not None and pointed != self.subject:
                    self.subject = pointed
                    break
        if self.subject is None:
            planes = self.canvas.fontLevelPlanesStack
            if planes:
                for plainName in planes.hitTestOrder:
                    plain = planes.get(plainName)
                    if plain.visible and not plain.locked:
                        pointed = plain.hitTest(canvas_x, canvas_y)
                        if pointed is not None and pointed != self.subject:
                            self.subject = pointed
                            break

    # Selection ------------------

    def isAnySelected(self) -> bool:
        glyph = self.glyph
        if glyph is not None:
            for contour in glyph:
                for point in contour:
                    if point.selected:
                        return True
            for anchor in glyph.anchors:
                if anchor.selected:
                    return True
            for component in glyph.components:
                if component.selected:
                    return True
            for guideline in glyph.guidelines:
                if guideline.selected:
                    return True
        for guideline in self.canvas.font.guidelines:
            if guideline.selected:
                return True
        return False

    def unselectAll(self) -> None:
        glyph = self.glyph
        if glyph is not None:
            for contour in glyph:
                contour.selected = False
            for anchor in glyph.anchors:
                anchor.selected = False
            for component in glyph.components:
                component.selected = False
            for guideline in glyph.guidelines:
                guideline.selected = False
        for guideline in self.canvas.font.guidelines:
            guideline.selected = False

    def postNotifications(self):
        # log.debug("postNotifications")
        # log.debug("activeStack       %s", self.activeStack)
        # log.debug("glyph             %s", self.glyph)
        # log.debug("contoursChanged   %s", self.contoursChanged)
        # log.debug("anchorsChanged    %s", self.anchorsChanged)
        # log.debug("componentsChanged %s", self.componentsChanged)
        # log.debug("glyphGuidelinesChanged %s", self.glyphGuidelinesChanged)
        # log.debug("fontGuidelinesChanged %s", self.fontGuidelinesChanged)
        glyph = self.glyph
        if glyph is not None:
            anyChanged = False
            glyph.disableNotifications()
            glyph.round(
                roundPoints=self.contoursChanged,
                roundAnchors=self.anchorsChanged,
                roundComponents=self.componentsChanged,
            )
            if self.glyphGuidelinesChanged:
                for guideline in glyph.guidelines:
                    guideline.round()
            glyph.enableNotifications()
            glyph.font.holdNotifications()
            if self.fontGuidelinesChanged:
                for guideline in glyph.font.guidelines:
                    guideline.round()
            if self.contoursChanged:
                for contour in glyph:
                    contour.destroyAllRepresentations()
                    contour.postNotification(notification="Contour.Changed")
                # componentReferences = glyph.font.componentReferences.get(glyph.name, ())
                # for compositeName in componentReferences:
                #     for component in glyph.font[compositeName].components:
                #         component.destroyAllRepresentations()
                glyph.postNotification(notification="Glyph.ContoursChanged")
                self.contoursChanged = False
                anyChanged = True
            if self.anchorsChanged:
                glyph.postNotification(notification="Glyph.AnchorsChanged")
                self.anchorsChanged = False
                anyChanged = True
            if self.componentsChanged:
                glyph.postNotification(notification="Glyph.ComponentsChanged")
                self.componentsChanged = False
                anyChanged = True
            if self.glyphGuidelinesChanged:
                glyph.postNotification(notification="Glyph.GuidelinesChanged")
                self.glyphGuidelinesChanged = False
                anyChanged = True
            if anyChanged:
                glyph.dirty = True
                glyph.postNotification(notification="Glyph.Changed")
            glyph.font.releaseHeldNotifications()

    def draw(self, gc):
        """
        may be overridden by subclass
        """

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_LEFT_DOWN(self, event):
        if not self.canvas.HasCapture():
            self.canvas.CaptureMouse()
        self.shiftDown = event.ShiftDown()
        if self.glyph is not None:
            self.xClick = self.x = event.GetX()
            self.yClick = self.y = event.GetY()

    def on_LEFT_UP(self, event):
        if self.canvas.HasCapture():
            self.canvas.ReleaseMouse()
        self.reset()
        glyph = self.glyph
        if glyph is not None:
            try:
                glyph.enableNotifications()
            except KeyError:
                pass
        self.canvas.Refresh()
        event.Skip()

    def on_RIGHT_DOWN(self, event):
        if self.state == IDLE:
            menu = DefaultEditMenu(self.canvas, self.subject)
            menu.x = self.canvas.screenToCanvasX(self.x)
            menu.y = self.canvas.screenToCanvasY(self.y)
            self.canvas.PopupMenu(menu, wx.DefaultPosition)
        else:
            event.Skip()

    def on_MOTION(self, event):
        self.shiftDown = event.ShiftDown()
        self.xPrev = self.x
        self.yPrev = self.y
        self.x = event.GetX()
        self.y = event.GetY()


class AddShapeTool(GlyphTool):
    """
    Base class for tools which add shapes like ellipsis or rectangles.
    """

    brush = wx.NullBrush
    pen = get_wxPen(wx.Colour(0, 0, 0, 255))

    @property
    def dimension(self):
        if self.glyph is not None:
            x = min(self.xClick, self.x)
            y = min(self.yClick, self.y)
            w = abs(self.xClick - self.x)
            h = abs(self.yClick - self.y)
            if self.shiftDown:
                w = h = max(w, h)
            return x, y, w, h
        return 0, 0, 0, 0

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_LEFT_DOWN(self, event):
        GlyphTool.on_LEFT_DOWN(self, event)
        if self.glyph is not None:
            self.state = RUBBERBANDING
        event.Skip()

    def on_MOTION(self, event):
        GlyphTool.on_MOTION(self, event)
        if event.Dragging() and self.state == RUBBERBANDING:
            x = min(self.xClick, self.xPrev)
            y = min(self.yClick, self.yPrev)
            w = abs(self.xClick - self.xPrev)
            h = abs(self.yClick - self.yPrev)
            rubberband = wx.Rect(x, y, w, h)
            rubberband.Union(wx.Rect(*self.dimension))
            rubberband.Inflate(1, 1)
            self.canvas.RefreshRect(rubberband)
        event.Skip()


class TransformBaseTool(GlyphTool):
    pen = get_wxPen(wx.Colour(128, 128, 128, 128))
    cursor = cursors["MarkCenter"]

    def __init__(self, parent):
        GlyphTool.__init__(self, parent)
        self.angleSnap = 15
        self.center = None
        self.ratio = None

    @property
    def angle(self):
        if self.state == WORKING:
            x = self.canvas.canvasToScreenX(self.center[0])
            y = self.canvas.canvasToScreenY(self.center[1])
            dx = x - self.x
            dy = y - self.y
            if dx == 0 and dy == 0:
                angle = 0
            elif dx > 0:
                angle = 180 - asin(dy / sqrt(dx**2 + dy**2)) * DEGREE
            else:
                angle = asin(dy / sqrt(dx**2 + dy**2)) * DEGREE
            if self.shiftDown:
                angle = round(angle / self.angleSnap) * self.angleSnap
            return -angle

    def draw(self, gc):
        if self.state == WORKING:
            d = 3
            x = self.canvas.canvasToScreenX(self.center[0])
            y = self.canvas.canvasToScreenY(self.center[1])
            gc.SetPen(self.pen)
            gc.StrokeLine(x - d, y - d, x + d, y + d)
            gc.StrokeLine(x + d, y - d, x - d, y + d)
            gc.StrokeLine(x, y, self.x, self.y)

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_LEFT_DOWN(self, event):
        GlyphTool.on_LEFT_DOWN(self, event)
        glyph = self.glyph
        if glyph is not None:
            self.center = (
                self.canvas.screenToCanvasX(self.xClick),
                self.canvas.screenToCanvasY(self.yClick),
            )
            glyph.undoManager.saveState()
            glyph.disableNotifications()
            self.state = WORKING
        event.Skip()

    def on_LEFT_UP(self, event):
        glyph = self.glyph
        if glyph is not None:
            glyph.enableNotifications()
            self.postNotifications()
        super().on_LEFT_UP(event)



"""
editPanel
===============================================================================
"""
import logging
import wx

from wbBase.tools import get_wxBrush, get_wxFont, get_wxPen

from .canvas import Canvas

log = logging.getLogger(__name__)

ID = wx.ID_ANY
pos = wx.DefaultPosition
size = wx.DefaultSize


class Ruler(wx.Panel):
    def __init__(self, parent, style=wx.NO_BORDER | wx.TAB_TRAVERSAL, name="Ruler"):
        super().__init__(parent, ID, pos, size, style, name)
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self.SetForegroundColour(wx.Colour(80, 80, 80))
        self.SetFont(get_wxFont(pointSize=6, faceName="Small Fonts"))
        self.createGuideline = None
        self.Bind(wx.EVT_ENTER_WINDOW, self.OnEnterWindow)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnLeaveWindow)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        self.Bind(wx.EVT_SIZE, self.on_SIZE)

    @property
    def canvas(self):
        return self.Parent.canvas

    @property
    def zoom(self):
        return self.canvas.zoom

    @property
    def tickSize(self):
        z = self.zoom
        if z <= 0.1:
            return 1000.0
        if z <= 0.2:
            return 500.0
        if z <= 0.5:
            return 200.0
        if z <= 1:
            return 100.0
        if z <= 2:
            return 50.0
        if z <= 5:
            return 10.0
        if z <= 20:
            return 5.0
        return 1.0

    def OnEnterWindow(self, event):
        self.createGuideline = None
        event.Skip()

    def OnLeaveWindow(self, event):
        if self.HasCapture():
            self.ReleaseMouse()
        print("", end=None) # magic - make it work on mac!?
        event.Skip()

    def OnLeftDown(self, event):
        if event.ShiftDown():
            self.createGuideline = "Font"
        else:
            self.createGuideline = "Glyph"
        event.Skip()

    def OnLeftUp(self, event):
        self.createGuideline = None
        event.Skip()

    def OnPaint(self, event):
        dc = wx.BufferedPaintDC(self)
        dc.SetBackground(get_wxBrush(self.BackgroundColour))
        dc.Clear()
        self.Draw(dc)

    def OnEraseBackground(self, event):
        pass

    def on_SIZE(self, event):
        self.Refresh()
        event.Skip()

    def Draw(self, dc):
        dc.SetPen(get_wxPen(self.ForegroundColour))
        dc.SetTextForeground(wx.BLACK)


class RulerHorizontal(Ruler):
    def __init__(self, parent):
        super().__init__(parent, name="RulerHorizontal")
        self.SetInitialSize((size.width, 20))

    @property
    def minValue(self):
        return self.canvas.screenToCanvasX(0)

    @property
    def maxValue(self):
        return self.canvas.screenToCanvasX(self.Size.width)

    def Draw(self, dc):
        super().Draw(dc)
        w = self.Size.width
        h = self.Size.height
        dc.DrawLine(0, h - 1, w, h - 1)
        ts = self.tickSize
        tick = int(round(self.minValue / ts) * ts)
        while tick < self.maxValue:
            t = self.canvas.canvasToScreenX(tick)
            dc.DrawLine(t, 0, t, h)
            dc.DrawText(str(int(tick)), t + 2, 1)
            tick += ts


class RulerVertical(Ruler):
    def __init__(self, parent):
        super().__init__(parent, name="RulerVertical")
        self.SetInitialSize((20, size.height))

    @property
    def minValue(self):
        return self.canvas.screenToCanvasY(self.Size.height)

    @property
    def maxValue(self):
        return self.canvas.screenToCanvasY(0)

    def Draw(self, dc):
        super().Draw(dc)
        w = self.Size.width - 1
        h = self.Size.height
        dc.DrawLine(w, 0, w, h)
        ts = self.tickSize
        tick = int(round(self.minValue / ts) * ts)
        while tick < self.maxValue:
            t = self.canvas.canvasToScreenY(tick)
            dc.DrawLine(0, t, w, t)
            text = str(int(tick))
            txt_w, txt_h = dc.GetTextExtent(text)
            dc.DrawText(text, w - txt_w - 1, t - txt_h)
            tick += ts


class GlyphEditPanel(wx.Panel):
    def __init__(self, parent, doc):
        super().__init__(
            parent,
            ID,
            pos,
            size,
            style=wx.NO_BORDER | wx.TAB_TRAVERSAL,
            name="UfoFontWindow",
        )
        self.SetBackgroundColour(wx.Colour(245, 245, 245))
        sizer = wx.FlexGridSizer(2, 2, 0, 0)
        sizer.AddGrowableCol(1)
        sizer.AddGrowableRow(1)
        sizer.SetFlexibleDirection(wx.BOTH)
        sizer.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_SPECIFIED)

        self.canvas = Canvas(self, doc=doc)
        self.rulerTop = RulerHorizontal(self)
        self.rulerLeft = RulerVertical(self)

        sizer.Add(0, 0)
        sizer.Add(self.rulerTop, 0, wx.EXPAND, 0)
        sizer.Add(self.rulerLeft, 0, wx.EXPAND, 0)
        sizer.Add(self.canvas, 1, wx.EXPAND, 0)
        self.SetSizer(sizer)
        self.Layout()

    def Destroy(self):
        wx.LogDebug("GlyphEditPanel.Destroy()")
        try:
            self.DestroyChildren()
        except wx.wxAssertionError:
            pass
        return super().Destroy()

    @property
    def rulerShown(self):
        return self.rulerTop.Shown or self.rulerLeft.Shown

    def showRuler(self, show=True):
        self.rulerTop.Show(show)
        self.rulerLeft.Show(show)
        self.SendSizeEvent()

    def hideRuler(self, hide=True):
        self.showRuler(not hide)



import logging
import re
from string import digits, whitespace
import wx
from fontTools.ufoLib import (
    fontInfoAttributesVersion3ValueData,
    validateFontInfoVersion3ValueForAttribute,
)

log = logging.getLogger(__name__)


class FontInfoBaseValidator(wx.Validator):
    def __init__(self, getFontInfo):
        super().__init__()
        self._getFontInfo = getFontInfo
        self._fontInfo = None
        self._valueType = None
        # log.debug("FontInfoBaseValidator.__init__() done")

    def __repr__(self):
        return "<%s for %r>" % (self.__class__.__name__, self.Window.Name)

    @property
    def fontInfo(self):
        if not self._fontInfo:
            self._fontInfo = self._getFontInfo()
        return self._fontInfo

    @property
    def valueType(self):
        if self._valueType is None:
            ctrl = self.GetWindow()
            if ctrl:
                self._valueType = fontInfoAttributesVersion3ValueData[ctrl.Name]["type"]
        return self._valueType

    def Clone(self):
        return self.__class__(self._getFontInfo)

    def Validate(self, win):
        return True

    def TransferToWindow(self):
        return True

    def TransferFromWindow(self):
        return True


class FontInfoTextCtrlValidator(FontInfoBaseValidator):
    def __init__(self, getFontInfo):
        super().__init__(getFontInfo)
        self.Bind(wx.EVT_CHAR, self.on_char)

    def _getValueFromControl(self):
        ctrl = self.GetWindow()
        value = ctrl.Value
        if value == wx.EmptyString:
            return None
        valueType = self.valueType
        if valueType == str:
            return value
        elif valueType == int:
            return int(value)
        elif valueType == float:
            return float(value)
        elif valueType in ((int, float), (float, int)):
            try:
                valInt = int(value)
            except ValueError:
                valInt = None
            try:
                valFloat = float(value)
            except ValueError:
                valFloat = None
            if valInt is None and valFloat is not None:
                return valFloat
            elif valInt is not None and valFloat is None:
                return valInt
            elif valInt is not None and valFloat is not None:
                if abs(valInt - valFloat) < 0.0001:
                    return valInt
                else:
                    return valFloat
        else:
            log.debug("valueType: %r, value: %r", valueType, value)
            raise ValueError(f"Can't get value '{ctrl.Name}' from control'")

    def Validate(self, win):
        ctrl = self.GetWindow()
        name = ctrl.Name
        # log.debug("Validate %r for %s", win, name)
        info = self.fontInfo
        if info:
            if hasattr(info, name):
                try:
                    value = self._getValueFromControl()
                except ValueError:
                    log.debug("ValueError during Validate for %s", name)
                    return False
                if not value:
                    return True
                result = validateFontInfoVersion3ValueForAttribute(name, value)
                if not result:
                    log.debug("Validate for for %s returned %r", name, result)
                return result
            else:
                log.debug("name not found in fontInfo during Validate for %s", name)
        else:
            log.debug("No fontInfo during Validate for %s", name)
        return True

    def TransferToWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info and hasattr(info, name):
            value = getattr(info, name)
            if value is None:
                ctrl.Value = wx.EmptyString
            elif isinstance(value, str):
                ctrl.Value = value
            else:
                try:
                    ctrl.Value = str(value)
                except ValueError:
                    return False
        return True
        # log.debug("TransferToWindow: no fontInfo")
        # return False

    def TransferFromWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info and hasattr(info, name):
            try:
                setattr(info, name, self._getValueFromControl())
                return True
            except ValueError:
                log.debug("ValueError during TransferFromWindow for %s", name)
                return False
        return False

    def on_char(self, event):
        key = event.GetKeyCode()
        if key < wx.WXK_SPACE or key == wx.WXK_DELETE or key > 255:
            event.Skip()
            return
        if self.valueType == str:
            event.Skip()
            return
        elif (
            self.valueType in (int, float, (int, float), (float, int))
            and chr(key) in digits + "+-."
        ):
            event.Skip()
            return
        if not wx.Validator.IsSilent():
            wx.Bell()
        return


class FontInfoIntlistTextCtrlValidator(FontInfoTextCtrlValidator):
    # def __init__(self, getFontInfo):
    #     super().__init__(getFontInfo)

    def _getValueFromControl(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info and hasattr(info, name):
            value = ctrl.Value
            if value == wx.EmptyString:
                return []
            else:
                return [int(v) for v in value.split()]
        raise ValueError(f"Can't get value '{name}' from control'")

    # def Validate(self, win):
    #     ctrl = self.GetWindow()
    #     name = ctrl.Name
    #     log.debug("Validate %r for %s", win, name)
    #     info = self.fontInfo
    #     if info:
    #         if hasattr(info, name):
    #             try:
    #                 value = self._getValueFromControl()
    #             except ValueError:
    #                 log.debug("ValueError during Validate for %s", name)
    #                 return False
    #             if not value :
    #                 return True
    #             result = validateFontInfoVersion3ValueForAttribute(name, value)
    #             log.debug("Validate for for %s returned %r", namer, esult)
    #             return result
    #         else:
    #             log.debug("name not found in fontInfo during Validate for %s", name)
    #     else:
    #         log.debug("No fontInfo during Validate for %s", name)
    #     return True

    def TransferToWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info and hasattr(info, name):
            value = getattr(info, name)
            if value is None:
                ctrl.Value = wx.EmptyString
            elif isinstance(value, (list, tuple)):
                if not value:
                    ctrl.Value = wx.EmptyString
                else:
                    ctrl.Value = " ".join([str(v) for v in value])
            else:
                return False
        return True

    def on_char(self, event):
        key = event.GetKeyCode()
        if key < wx.WXK_SPACE or key == wx.WXK_DELETE or key > 255:
            event.Skip()
            return
        if chr(key) in digits + whitespace + "+-":
            event.Skip()
            return
        if not wx.Validator.IsSilent():
            wx.Bell()
        return


class FontInfoPShintCtrlValidator(FontInfoBaseValidator):

    def _getValueFromControl(self):
        ctrl = self.GetWindow()
        value = ctrl.Value
        if value:
            return value

    def TransferToWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info and hasattr(info, name):
            value = getattr(info, name)
            if value is None:
                ctrl.Value = []
            elif isinstance(value, (list, tuple)):
                ctrl.Value = value
            else:
                return False
        return True

    def TransferFromWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info and hasattr(info, name):
            oldValue = getattr(info, name)
            newValue = self._getValueFromControl()
            if newValue != oldValue:
                try:
                    setattr(info, name, newValue)
                    return True
                except ValueError:
                    log.debug("ValueError during TransferFromWindow for %s", name)
                    return False
        return False


class FontInfoCheckBoxValidator(FontInfoBaseValidator):
    def __init__(self, getFontInfo):
        super().__init__(getFontInfo)
        self.flagsPattern = re.compile(r"^\w+ \d{1,2}$")

    @property
    def valueType(self):
        if self._valueType is None:
            ctrl = self.GetWindow()
            if ctrl:
                name = ctrl.Name
                if self.flagsPattern.match(name):
                    attribute, __ = name.split(" ", 1)
                else:
                    attribute = name
                self._valueType = fontInfoAttributesVersion3ValueData[attribute]["type"]
        return self._valueType

    def TransferToWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info:
            if self.flagsPattern.match(name):
                attribute, value = name.split(" ", 1)
                value = int(value)
                if hasattr(info, attribute):
                    # log.debug(
                    #     "FontInfoCheckBoxValidator.TransferToWindow(%r)",
                    #     getattr(info, attribute),
                    # )
                    infoValue = getattr(info, attribute)
                    if infoValue:
                        ctrl.Value = value in infoValue
                    else:
                        ctrl.Value = False
                else:
                    ctrl.Value = False
            else:
                attribute = name
                if hasattr(info, attribute):
                    ctrl.Value = bool(getattr(info, attribute))
                else:
                    ctrl.Value = False
        else:
            ctrl.Value = False
        return True

    def TransferFromWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info:
            if self.flagsPattern.match(name):
                attribute, value = name.split(" ", 1)
                value = int(value)
                infoValue = list(getattr(info, attribute))
                if ctrl.Value and value not in infoValue:
                    infoValue.append(value)
                    infoValue.sort()
                elif not ctrl.Value and value in infoValue:
                    infoValue.remove(value)
            else:
                attribute = name
                infoValue = ctrl.Value
            log.debug("TransferFromWindow %r, %r", attribute, infoValue)
            setattr(info, attribute, infoValue)
            return True
        log.debug("ValueError during TransferFromWindow for %s", name)
        return False


class FontInfoBitListValidator(FontInfoBaseValidator):
    def TransferToWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info and hasattr(info, name):
            ctrl.Value = getattr(info, name)
        else:
            ctrl.Value = ()
        return True

    def TransferFromWindow(self):
        ctrl = self.GetWindow()
        name = ctrl.Name
        info = self.fontInfo
        if info and hasattr(info, name):
            setattr(info, name, ctrl.Value)
            return True
        log.debug("ValueError during TransferFromWindow for %s", name)
        return False



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version Oct 26 2018)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc

###########################################################################
## Class GenerateOTCFFoptionsDialogUI
###########################################################################

class GenerateOTCFFoptionsDialogUI ( wx.Dialog ):

    def __init__( self, parent ):
        wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = u"Generate OT-CFF", pos = wx.DefaultPosition, size = wx.DefaultSize, style = wx.CAPTION|wx.RESIZE_BORDER )

        self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )

        sizer = wx.BoxSizer( wx.VERTICAL )

        self.message = wx.StaticText( self, wx.ID_ANY, u"Options", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.message.Wrap( -1 )

        sizer.Add( self.message, 0, wx.ALL, 5 )

        gbSizer = wx.GridBagSizer( 0, 0 )
        gbSizer.SetFlexibleDirection( wx.BOTH )
        gbSizer.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

        self.checkBox_autohint = wx.CheckBox( self, wx.ID_ANY, u"PS-Autohint", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.checkBox_autohint.SetValue(True)
        gbSizer.Add( self.checkBox_autohint, wx.GBPosition( 0, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_removeOverlap = wx.CheckBox( self, wx.ID_ANY, u"Remove Overlap", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.checkBox_removeOverlap.SetValue(True)
        gbSizer.Add( self.checkBox_removeOverlap, wx.GBPosition( 1, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.staticText_cffcompress = wx.StaticText( self, wx.ID_ANY, u"CFF Compress", wx.DefaultPosition, wx.DefaultSize, 0 )
        self.staticText_cffcompress.Wrap( -1 )

        gbSizer.Add( self.staticText_cffcompress, wx.GBPosition( 2, 0 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        choice_cffcompressChoices = [ u"NONE", u"SPECIALIZE", u"SUBROUTINIZE" ]
        self.choice_cffcompress = wx.Choice( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, choice_cffcompressChoices, 0 )
        self.choice_cffcompress.SetSelection( 2 )
        gbSizer.Add( self.choice_cffcompress, wx.GBPosition( 2, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_useProductionNames = wx.CheckBox( self, wx.ID_ANY, u"Use Production Names", wx.DefaultPosition, wx.DefaultSize, 0 )
        gbSizer.Add( self.checkBox_useProductionNames, wx.GBPosition( 3, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_writeKernFeature = wx.CheckBox( self, wx.ID_ANY, u"Write kern Feature", wx.DefaultPosition, wx.DefaultSize, 0 )
        gbSizer.Add( self.checkBox_writeKernFeature, wx.GBPosition( 4, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

        self.checkBox_writeMarkFeature = wx.CheckBox( self, wx.ID_ANY, u"Write mark Feature", wx.DefaultPosition, wx.DefaultSize, 0 )
        gbSizer.Add( self.checkBox_writeMarkFeature, wx.GBPosition( 5, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )


        sizer.Add( gbSizer, 1, wx.ALL|wx.EXPAND, 5 )

        self.staticline = wx.StaticLine( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.LI_HORIZONTAL )
        sizer.Add( self.staticline, 0, wx.BOTTOM|wx.EXPAND|wx.TOP, 5 )

        sdbSizer = wx.StdDialogButtonSizer()
        self.sdbSizerOK = wx.Button( self, wx.ID_OK )
        sdbSizer.AddButton( self.sdbSizerOK )
        self.sdbSizerCancel = wx.Button( self, wx.ID_CANCEL )
        sdbSizer.AddButton( self.sdbSizerCancel )
        sdbSizer.Realize();

        sizer.Add( sdbSizer, 0, wx.ALL|wx.EXPAND, 5 )


        self.SetSizer( sizer )
        self.Layout()
        sizer.Fit( self )

        self.Centre( wx.BOTH )

    def __del__( self ):
        pass





from __future__ import annotations
import logging
from typing import Optional, Tuple, TYPE_CHECKING

import wx
import wx.grid as gridlib
from wbDefcon import Font
from wbBase.scripting import MacroButtonMixin

from .kerningStatusPanelUI import KerningStatusPanelUI

if TYPE_CHECKING:
    from wbDefcon import Kerning
    from ..fontinfo import UfoFontInfoWindow

log = logging.getLogger(__name__)


class KerningStatusPanel(KerningStatusPanelUI, MacroButtonMixin):
    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE,
    ):
        super().__init__(parent, id, pos, size, style, name="KerningStatusPanel")
        MacroButtonMixin.__init__(
            self, self.button_macro, "_kerning", self.GrandParent.view
        )

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_button_New(self, event):
        event.Skip()

    def on_button_Delete(self, event):
        event.Skip()

    def update_button_Delete(self, event):
        event.Skip()


class KerningTable(gridlib.GridTableBase):
    def __init__(self, kerning: Kerning):
        super().__init__()
        self.kerning = kerning
        self.colLabels = ("Side 1", "Side 2", "Value")
        self.dataTypes = (
            gridlib.GRID_VALUE_STRING,
            gridlib.GRID_VALUE_STRING,
            gridlib.GRID_VALUE_NUMBER,
        )

    def GetNumberCols(self) -> int:
        return 3

    def GetNumberRows(self) -> int:
        return len(self.kerning)

    def GetColLabelValue(self, col: int) -> str:
        return self.colLabels[col]

    def GetRowLabelValue(self, row: int) -> str:
        return ""

    def CanGetValueAs(self, row: int, col: int, typeName) -> bool:
        colType = self.dataTypes[col].split(":")[0]
        if typeName == colType:
            return True
        else:
            return False

    def CanSetValueAs(self, row: int, col: int, typeName):
        return self.CanGetValueAs(row, col, typeName)

    def GetValue(self, row: int, col: int):
        pair: Tuple[str, str] = self.kerning.sortedPairList[row]
        if col == 0:
            return pair[0].replace("public.kern1.", "@", 1)
        if col == 1:
            return pair[1].replace("public.kern2.", "@", 1)
        if col == 2:
            return self.kerning[pair]
        # return ("A", "V", -10)[col]

    def SetValue(self, row: int, col: int, value):
        if col == 2:
            kernValue = int(value)
            pair = self.kerning.sortedPairList[row]
            self.kerning[pair] = kernValue

    def GetAttr(self, row: int, col: int, kind) -> gridlib.GridCellAttr:
        attr = gridlib.GridCellAttr()
        if row % 2:
            attr.SetBackgroundColour(wx.Colour("WHITESMOKE"))
        else:
            attr.SetBackgroundColour(wx.Colour("WHITE"))
        if col in (0, 1):
            attr.SetReadOnly()
            if self.GetValue(row, col).startswith("@"):
                attr.SetTextColour("MEDIUM BLUE")
        elif col == 2:
            attr.SetAlignment(wx.ALIGN_RIGHT, wx.ALIGN_CENTER)
        return attr

    def IsEmptyCell(self, row: int, col: int) -> bool:
        return False


class KerningGrid(gridlib.Grid):
    Table: KerningTable

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.WANTS_CHARS | wx.NO_BORDER | gridlib.Grid.GridSelectRows,
    ):
        super().__init__(parent, id, pos, size, style, name="KerningGrid")
        self.SetColLabelSize(20)
        self.SetRowLabelSize(0)
        # self.UseNativeColHeader(True)
        self.SetDefaultCellFitMode(gridlib.GridFitMode.Ellipsize(wx.ELLIPSIZE_MIDDLE))
        # self.SetSortingColumn(0)
        # bind events
        self.Bind(gridlib.EVT_GRID_SELECT_CELL, self.on_selectCell)

    @property
    def kerning(self):
        return self.Table.kerning

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_selectCell(self, event):
        pair = self.kerning.sortedPairList[event.GetRow()]
        log.debug("%r: %r", pair, self.kerning[pair])
        event.Skip()


class KerningPage(wx.Panel):
    Parent: UfoFontInfoWindow

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.TAB_TRAVERSAL,
    ):
        super().__init__(parent, id, pos, size, style, name="KerningPage")
        self._font: Optional[Font] = None
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.kerningPanel = wx.SplitterWindow(
            self,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SP_3D | wx.SP_NOBORDER | wx.BORDER_NONE,
        )
        self.kerningGrid = KerningGrid(self.kerningPanel)
        self.kerningViewPanel = wx.Panel(self.kerningPanel)
        sizer.Add(self.kerningPanel, 1, wx.EXPAND, 0)
        self.kerningPanel.SetMinimumPaneSize(20)
        self.kerningPanel.SashGravity = 0.3
        self.kerningPanel.SplitVertically(self.kerningGrid, self.kerningViewPanel, 200)

        self.kerningStatusPanel = KerningStatusPanel(self)
        sizer.Add(self.kerningStatusPanel, 0, wx.EXPAND, 0)
        self.SetSizer(sizer)
        self.Layout()
        # Connect Events
        self.kerningPanel.Bind(wx.EVT_IDLE, self.kerningPanelOnIdle)

    @property
    def font(self) -> Optional[Font]:
        """The wbDefcon Font object of this KerningPage"""
        return self._font

    @font.setter
    def font(self, value: Font) -> None:
        assert self._font is None
        assert isinstance(value, Font)
        assert value == self.Parent.font
        self._font = value
        self.kerningGrid.SetTable(
            KerningTable(self._font.kerning), True, self.kerningGrid.GridSelectRows
        )
        self._font.kerning.addObserver(
            self,
            "handleNotification",
            notification=self._font.kerning.changeNotificationName,
        )
        # self.kerningGrid.SetSortingColumn(0)

    @font.deleter
    def font(self) -> None:
        if self._font is not None and self._font.kerning.hasObserver(
            self, self._font.kerning.changeNotificationName
        ):
            self._font.kerning.removeObserver(
                self, self._font.kerning.changeNotificationName
            )
        self._font = None

    def handleNotification(self, notification):
        self.Refresh()

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def kerningPanelOnIdle(self, event):
        self.kerningPanel.SetSashPosition(200)
        self.kerningPanel.Unbind(wx.EVT_IDLE)



"""
revertFontDialog
===============================================================================

This module provides the dialog which allows selective loading of external changes.
"""
from __future__ import annotations
from typing import TYPE_CHECKING
import logging
import wx

from .revertFontDialogUI import RevertFontDialogUI, LayerPanelUI

if TYPE_CHECKING:
    from wbDefcon import Font

log = logging.getLogger(__name__)

class LayerPanel(LayerPanelUI):
    """
    Component for RevertFontDialog
    """
    def __init__(self, parent, name, info=False, glyphNames=None):
        super().__init__(
            parent,
            id=wx.ID_ANY,
            pos=wx.DefaultPosition,
            size=wx.DefaultSize,
            style=wx.TAB_TRAVERSAL,
            name=name,
        )
        self.lbl_name.LabelText = f"Layer: {name}"
        self.checkBox_info.Value = info
        if glyphNames:
            assert isinstance(glyphNames, (list, tuple))
            self.checkList_glyphNames.SetItems(glyphNames)
            self.checkList_glyphNames.SetCheckedStrings(glyphNames)


class RevertFontDialog(RevertFontDialogUI):
    def __init__(self, parent, font:Font):
        super().__init__(parent)
        externalChanges = font.testForExternalChanges()
        log.debug("externalChanges: %r", externalChanges)
        self.checkBox_font_lib.Value = font.lib.dirty or externalChanges["lib"]
        self.checkBox_font_info.Value = font.info.dirty or externalChanges["info"]
        self.checkBox_groups.Value = font.groups.dirty or externalChanges["groups"]
        self.checkBox_kerning.Value = font.kerning.dirty or externalChanges["kerning"]
        self.checkBox_features.Value = (
            font.features.dirty or externalChanges["features"] or False
        )
        self.checkBox_images.Value = font.images.dirty
        self.checkBox_data.Value = font.data.dirty
        self.checkBox_defaultLayer.Value = externalChanges["layers"]["defaultLayer"]
        self.checkBox_layer_order.Value = externalChanges["layers"]["order"]
        self.layerPanels = []
        modifiedGlyphs = {}
        for layer in font.layers:
            for glyph in layer:
                if glyph.dirty:
                    if layer.name not in modifiedGlyphs:
                        modifiedGlyphs[layer.name] = []
                    modifiedGlyphs[layer.name].append(glyph.name)
        if modifiedGlyphs or externalChanges["layers"]["modified"]:
            sizer_layers_list = wx.BoxSizer(wx.VERTICAL)
            for layer in font.layers:
                externalModifiedGlyphs = []
                externalModifiedInfo = False
                if layer.name in externalChanges["layers"]["modified"]:
                    externalModifiedGlyphs = externalChanges["layers"]["modified"][
                        layer.name
                    ]["modified"]
                    externalModifiedInfo = externalChanges["layers"]["modified"][
                        layer.name
                    ]["info"]
                locallModifiedGlyphs = modifiedGlyphs.get(layer.name, [])
                if externalModifiedGlyphs or locallModifiedGlyphs:
                    allModifiedGlyphs = externalModifiedGlyphs + [
                        g
                        for g in locallModifiedGlyphs
                        if g not in externalModifiedGlyphs
                    ]
                    layerPanel = LayerPanel(
                        self.scrolledWindow_layers,
                        layer.name,
                        info=externalModifiedInfo,
                        glyphNames=allModifiedGlyphs,
                    )
                    sizer_layers_list.Add(
                        layerPanel, 0, wx.BOTTOM | wx.EXPAND | wx.TOP, 5
                    )
                    self.layerPanels.append(layerPanel)

            self.scrolledWindow_layers.SetSizer(sizer_layers_list)
            self.scrolledWindow_layers.Layout()
            sizer_layers_list.Fit(self.scrolledWindow_layers)
        self.Layout()

    @property
    def layerdata(self):
        data = dict(
            order=self.checkBox_layer_order.Value,
            default=self.checkBox_defaultLayer.Value,
        )
        if self.layerPanels:
            layers = {}
            for layerPanel in self.layerPanels:
                layers[layerPanel.Name] = dict(
                    info=layerPanel.checkBox_info.Value,
                    glyphNames=layerPanel.checkList_glyphNames.CheckedStrings,
                )
            data["layers"] = layers
        return data

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_checkBox_all(self, event):
        for checkBox in (
            self.checkBox_font_lib,
            self.checkBox_font_info,
            self.checkBox_groups,
            self.checkBox_kerning,
            self.checkBox_features,
            self.checkBox_images,
            self.checkBox_data,
            self.checkBox_defaultLayer,
            self.checkBox_layer_order,
        ):
            checkBox.Value = event.IsChecked()
        event.Skip()

    def on_checkBox_single(self, event):
        if not event.IsChecked():
            self.checkBox_all.Value = False
        event.Skip()



"""
newFontDialog
===============================================================================
"""
import wx

from .newFontDialogUI import NewFontDialogUI


class NewFontDialog(NewFontDialogUI):
    @property
    def familyName(self):
        return self.textCtrl_family.Value

    @property
    def styleName(self):
        return self.textCtrl_style.Value

    @property
    def unitsPerEm(self):
        return self.spinCtrl_upm.Value



"""
editTool
===============================================================================

Implementation of the Glyph Edit Toolbar
"""
import logging
import time

import wx
from wx import aui

from ..dialog.layerDialog import getLayerInfo
from ..view.glyph import UfoGlyphView
from . import BaseToolbar

log = logging.getLogger(__name__)

(
    edit,
    erase,
    knife,
    meter,
    startpoint,
    reversecontour,
    draw,
    rotate,
    scale,
    rectangle,
    ellipse,
    layer,
) = range(12)


class GlyphEditToolbar(BaseToolbar):
    """Glyph Edit Toolbar"""

    itemType = wx.ITEM_RADIO
    tools = {
        edit: "ToolEdit",
        erase: "ToolErase",
        knife: "ToolKnife",
        meter: "ToolMeter",
        startpoint: "ToolStartpoint",
        reversecontour: "ToolReverse",
        draw: "ToolDraw",
        rectangle: "AddRectangle",
        ellipse: "AddEllipse",
        rotate: "ToolRotate",
        scale: "ToolScale",
        # "ToolZoom",
        # "ToolPan",
        # "MagicWand",
    }

    def __init__(self, parent):
        super().__init__(parent, "GlyphEditToolbar")
        self.appendTool("Edit", "TOOL_EDIT", commandIndex=edit)
        self.appendTool("Erase (Q)", "ERASER", commandIndex=erase)
        self.appendTool("Knife (W)", "TOOL_KNIFE", commandIndex=knife)
        self.appendTool("Meter (T)", "TOOL_METER", commandIndex=meter)
        self.appendTool("Startpoint", "SET_STARTPOINT", commandIndex=startpoint)
        self.appendTool("Reverse", "REVERSE_CONTOUR", commandIndex=reversecontour)
        self.appendTool("Draw (D)", "TOOL_DRAW", commandIndex=draw)
        self.appendTool("Rectangle (R)", "TOOL_RECTANGLE", commandIndex=rectangle)
        self.appendTool("Ellipse (E)", "TOOL_ELLIPSE", commandIndex=ellipse)
        self.appendTool("Rotate", "TOOL_ROTATE", commandIndex=rotate)
        self.appendTool("Scale", "TOOL_SCALE", commandIndex=scale)
        self.AddSeparator()
        layer_tool = self.appendTool(
            "Layer", "LAYER_EDIT", "Edit Layer", layer, wx.ITEM_NORMAL
        )
        self.SetToolDropDown(layer_tool.Id, True)
        self.Bind(aui.EVT_AUITOOLBAR_TOOL_DROPDOWN, self.on_Tool, layer_tool)

    def on_activate_layer(self, event):
        layerName = event.EventObject.FindItemById(event.Id).ItemLabelText
        view = self.app.documentManager.currentView
        if layerName == view.frame.canvas.getActiveDrawingPlaneStack().name:
            return
        font = view.font
        layerColor = None
        if layerName == "<New Layer>":
            defaultName = "public.background"
            if defaultName in font.layers:
                defaultName = time.strftime("save %Y.%m.%d-%H:%M")
            layerName, layerColor = getLayerInfo(defaultName)
            if not layerName:
                return
        glyph = view.frame.canvas.glyph
        if layerName in font.layers:
            layer = font.layers[layerName]
        else:
            layer = font.newLayer(layerName)
        if layerColor:
            layer.color = layerColor
        if glyph.name not in layer:
            newGlyph = layer.newGlyph(glyph.name)
            newGlyph.width = glyph.width
        for layerPlane in view.frame.canvas.layerPlanes:
            layerPlane.active = layerPlane.name == layerName
        view.frame.canvas.glyph = layer[glyph.name]
        view.frame.SetFocus()
        event.Skip()

    def on_Tool(self, event):
        tool = self.FindTool(event.Id)
        i = tool.GetUserData()
        view = self.app.documentManager.currentView
        if i == layer:
            self.SetToolSticky(event.Id, True)
            menu = wx.Menu()
            activeLayerName = view.frame.canvas.getActiveDrawingPlaneStack().name
            for _layer in view.font.layers:
                name = _layer.name
                mnuItem = menu.Append(wx.ID_ANY, name, name, wx.ITEM_RADIO)
                mnuItem.Check(name == activeLayerName)
                self.Bind(wx.EVT_MENU, self.on_activate_layer, mnuItem)
            menu.AppendSeparator()
            mnuItem = menu.Append(
                wx.ID_ANY, "<New Layer>", "Switch to New Layer", wx.ITEM_NORMAL
            )
            self.Bind(wx.EVT_MENU, self.on_activate_layer, mnuItem)
            self.SetToolSticky(event.Id, False)
            self.PopupMenu(menu)
            menu.Destroy()
        elif isinstance(i, int) and i >= 0:
            view.frame.canvas.selectTool(self.tools[i])
        event.Skip()

    def on_update_Tool(self, event):
        view = self.app.documentManager.currentView
        if view and view.document:
            if isinstance(view, UfoGlyphView):
                event.Enable(True)
            else:
                event.Enable(False)
        else:
            event.Enable(False)



from __future__ import annotations

import logging
import pickle
from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

import wx
from fontTools.misc.transform import Transform
from fontTools.pens.transformPen import TransformPen
from wbDefcon import Font, Glyph
from wbDefcon.pens import graphicsPen
from wbFontParts import RGlyph
from wbBase.tools import get_wxBrush

from ..document import UfoDocument

if TYPE_CHECKING:
    from wbBase.application import App
    from wbDefcon import Info
    from .glyphViewPanel import GlyphViewPanel

    DragResult = int

log = logging.getLogger(__name__)


class GlyphViewCanvasDropTarget(wx.DropTarget):
    """
    DropTarget to handle dropped glyphs
    """

    DataObject: wx.CustomDataObject

    def __init__(self, canvas: GlyphViewCanvas):
        super().__init__()
        self.canvas: GlyphViewCanvas = canvas
        self.SetDataObject(wx.CustomDataObject("application.UFO-WB.glyphNames"))
        self.SetDefaultAction(wx.DragCopy)

    @property
    def app(self) -> App:
        return wx.GetApp()

    def OnEnter(self, x, y, defResult: DragResult):
        self.canvas.SetFocus()
        return defResult

    def OnDragOver(self, x, y, defResult: DragResult):
        self.canvas.setCaretPosition(x, y)
        return defResult

    def OnDrop(self, x: int, y: int) -> bool:
        return True

    def OnData(self, x, y, defResult: DragResult) -> DragResult:
        if self.GetData():
            glyphNameData = self.DataObject.GetData()
            if glyphNameData:
                glyphNameDict = pickle.loads(glyphNameData)
                font = None
                for doc in [
                    d
                    for d in self.app.documentManager.documents
                    if isinstance(d, UfoDocument)
                ]:
                    if id(doc.font) == glyphNameDict["font"]:
                        font = doc.font
                        break
                if not font:
                    return wx.DragError
                for glyph in [font[g] for g in glyphNameDict["glyphs"] if g in font]:
                    self.canvas.insertGlyphAtCaret(glyph)
                return defResult
        return wx.DragNone


class GlyphViewCanvas(wx.ScrolledWindow):
    """
    Canvas to draw glyphs on.
    """

    Parent: GlyphViewPanel
    refreshNotifications = (
        "Glyph.Changed",
        "Glyph.WidthChanged",
        "Glyph.ContoursChanged",
        "Glyph.ComponentsChanged",
        "Glyph.ComponentBaseGlyphDataChanged",
    )

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.ALWAYS_SHOW_SB | wx.HSCROLL | wx.VSCROLL,
    ):
        super().__init__(parent, id=id, pos=pos, size=size, style=style)
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self.SetBackgroundColour("WHITE")
        self.SetDropTarget(GlyphViewCanvasDropTarget(self))
        self._glyphs: List[Union[str, Glyph]] = []
        self.glyphPositions: List[Tuple[int, int]] = [(0, self.linespace)]

        # set caret
        self.Caret = wx.Caret(self, 2, self.linespace)
        self._caretIndex: int = 0
        self.Caret.Move(self.glyphPositions[self.caretIndex])
        self.Caret.Show()

        # Connect Events
        # window events
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SCROLLWIN, self.on_SCROLLWIN)
        # keyboard events
        self.Bind(wx.EVT_KEY_DOWN, self.on_KEY_DOWN)
        self.Bind(wx.EVT_CHAR, self.on_CHAR)

        # mouse events
        self.Bind(wx.EVT_LEFT_DOWN, self.on_LEFT_DOWN)
        self.Bind(wx.EVT_LEFT_UP, self.on_LEFT_UP)
        self.Bind(wx.EVT_LEFT_DCLICK, self.on_LEFT_DCLICK)
        self.Bind(wx.EVT_MOUSEWHEEL, self.on_MOUSEWHEEL)

    @property
    def app(self) -> App:
        return wx.GetApp()

    @property
    def fontsize(self) -> int:
        return round(self.Parent.fontsize)

    @fontsize.setter
    def fontsize(self, value: int):
        self.Parent.fontsize = int(value)

    @property
    def linespace(self) -> int:
        return round(self.Parent.fontsize * self.Parent.linespace / 100)

    @property
    def descender(self) -> int:
        return self.linespace - self.fontsize

    @property
    def showKerning(self) -> bool:
        return self.Parent.checkBox_showKerning.Value

    @property
    def glyphs(self) -> Iterator[Union[str, Glyph]]:
        self.checkGlyphs()
        for glyph in self._glyphs:
            yield glyph

    @property
    def glyphCount(self) -> int:
        """
        :return: Number of glyphs shown on the canvas, incl. line breaks
        """
        self.checkGlyphs()
        return len(self._glyphs)

    @property
    def caretIndex(self) -> int:
        """
        :return: Index of the current cursor position
        """
        return self._caretIndex

    @caretIndex.setter
    def caretIndex(self, value: int) -> None:
        self._caretIndex = abs(int(value))
        self.Refresh()

    @property
    def currentFont(self) -> Optional[Font]:
        """
        :return: Font at the current cursor position
        """
        self.checkGlyphs()
        if not self._glyphs:
            doc = self.app.documentManager.currentDocument
            if isinstance(doc, UfoDocument):
                return doc.font
            return None
        glyph = self._glyphs[max(self._caretIndex - 1, 0)]
        if self._caretIndex == 0 and isinstance(glyph, Glyph):
            return glyph.font
        for index in reversed(range(self._caretIndex)):
            glyph = self._glyphs[index]
            if isinstance(glyph, Glyph):
                return glyph.font
        return None

    @property
    def fonts(self) -> Tuple[Font]:
        self.checkGlyphs()
        _fonts = []
        for glyph in self._glyphs:
            if (
                isinstance(glyph, Glyph)
                and glyph.font is not None
                and glyph.font not in _fonts
            ):
                _fonts.append(glyph.font)
        return tuple(_fonts)

    def insertGlyph(self, index: int, glyph: Union[str, Glyph]):
        _glyph = None
        if isinstance(glyph, str) and glyph in ("\n", "\r"):
            _glyph = "\n"
        elif isinstance(glyph, Glyph):
            _glyph = glyph
        elif isinstance(glyph, RGlyph):
            _glyph = glyph.naked()
        if _glyph is not None:
            if isinstance(_glyph, Glyph):
                for notificationName in self.refreshNotifications:
                    if not _glyph.hasObserver(self, notificationName):
                        _glyph.addObserver(self, "handleNotification", notificationName)
                if _glyph.font not in self.fonts:
                    _glyph.font.kerning.addObserver(
                        self, "handleNotification", "Kerning.Changed"
                    )
            self._glyphs.insert(index, _glyph)
            self.caretIndex = index + 1
            self.SetFocus()

    def appendGlyph(self, glyph: Union[str, Glyph]):
        self.insertGlyph(self.glyphCount, glyph)

    def insertGlyphAtCaret(self, glyph: Union[str, Glyph]):
        self.insertGlyph(self.caretIndex, glyph)

    def checkGlyphs(self):
        """
        Remove all glyphs from fonts which have been closed
        """
        allFonts = [
            d.font
            for d in self.app.documentManager.documents
            if isinstance(d, UfoDocument)
        ]
        glyphs = []
        for glyph in self._glyphs:
            if isinstance(glyph, str):
                glyphs.append(glyph)
            elif glyph.font in allFonts:
                glyphs.append(glyph)
        if self._glyphs != glyphs:
            self._glyphs = glyphs
            self._caretIndex = 0

    def clearGlyphs(self) -> None:
        for font in self.fonts:
            font.kerning.removeObserver(self, "Kerning.Changed")
        for glyph in self._glyphs:
            if isinstance(glyph, str):
                continue
            for notificationName in self.refreshNotifications:
                glyph.removeObserver(self, notificationName)
        self._glyphs = []
        self.caretIndex = 0

    def popGlyph(self, index):
        if 0 <= index < self.glyphCount:
            glyph = self._glyphs.pop(index)
            if isinstance(glyph, Glyph) and glyph not in self._glyphs:
                for notificationName in self.refreshNotifications:
                    glyph.removeObserver(self, notificationName)
                if glyph.font not in self.fonts:
                    glyph.font.kerning.removeObserver(self, "Kerning.Changed")
            if self.caretIndex >= index:
                self.caretIndex -= 1
            self.Refresh()
            self.SetFocus()
            return glyph
        return None

    def setCaretPosition(self, x: int, y: int) -> None:
        if not self._glyphs:
            self.caretIndex = 0
            return
        self.checkGlyphs()
        view_x, view_y = self.GetViewStart()
        position = wx.Point(x + view_x, y + view_y)
        prev_x = None
        prev_y = None
        for i, (px, py) in enumerate(self.glyphPositions):
            if prev_x is not None and prev_y is not None:
                glyphRect = wx.Rect(
                    round(prev_x),
                    round(py - self.fontsize),
                    round(px - prev_x),
                    round(self.linespace),
                )
                if glyphRect.Top > position.y:
                    self.caretIndex = i - 1
                    return
                if glyphRect.Contains(position):
                    self.caretIndex = i - 1
                    return
            prev_x = px
            prev_y = py
        self.caretIndex = self.glyphCount

    def handleNotification(self, notification):
        log.debug("handleNotification: %r", notification)
        self.Refresh()

    def Refresh(self) -> None:
        self.Caret.SetSize(2, self.linespace)
        super().Refresh()

    # =================================================================
    # event handlers
    # =================================================================

    def on_SCROLLWIN(self, event):
        self.Refresh()
        event.Skip()

    def OnPaint(self, event: wx.PaintEvent) -> None:
        self.checkGlyphs()
        width, height = self.ClientSize
        current_x = 0
        current_y = self.fontsize
        self.glyphPositions.clear()
        self.glyphPositions.append((current_x, current_y))
        prev_glyph = None
        scale = 1
        for glyph in self._glyphs:
            if isinstance(glyph, Glyph) and glyph.font is not None:
                font = glyph.font
                if isinstance(font, Font):
                    info: Info = font.info
                    scale = self.fontsize / info.unitsPerEm
                    if (
                        isinstance(prev_glyph, Glyph)
                        and self.showKerning
                        and font == prev_glyph.font
                    ):
                        kernValue = font.kerning.find((prev_glyph.name, glyph.name))
                        if kernValue:
                            prev_x, prev_y = self.glyphPositions[-1]
                            prev_x += kernValue * scale
                            self.glyphPositions[-1] = (prev_x, prev_y)
                            current_x += round(kernValue * scale)
                glyphWidth = glyph.width * scale
                current_x += glyphWidth
            elif glyph == "\n":
                current_x = 0
                current_y += self.linespace
                prev_glyph = None

            width = max(width, current_x)
            height = max(height, current_y + self.descender)
            self.glyphPositions.append((current_x, current_y))
            prev_glyph = glyph
        self.VirtualSize = (width, height)
        caret_x, caret_y = self.glyphPositions[self.caretIndex]
        caret_y -= self.fontsize
        view_x, view_y = self.GetViewStart()
        caret_x -= view_x
        caret_y -= view_y
        self.Caret.Move(round(caret_x), round(caret_y))
        bitmap = wx.Bitmap(self.VirtualSize.width, self.VirtualSize.height)
        dc = wx.BufferedPaintDC(self, bitmap, wx.BUFFER_VIRTUAL_AREA)
        self.OnDraw(dc)

    def OnDraw(self, dc: wx.BufferedPaintDC) -> None:
        dc.SetBackground(get_wxBrush(self.GetBackgroundColour()))
        dc.Clear()
        gc: wx.GraphicsContext
        gc = wx.GraphicsContext.Create(dc)
        gc.SetBrush(wx.BLACK_BRUSH)
        gc.SetPen(wx.TRANSPARENT_PEN)
        for i, glyph in enumerate(self.glyphs):
            if isinstance(glyph, Glyph):
                scale = 1
                font = glyph.font
                if isinstance(font, Font):
                    scale = self.fontsize / font.info.unitsPerEm
                graphicsPen.path = gc.CreatePath()
                graphicsPen.glyphSet = glyph.layer
                transform = (
                    Transform().translate(*self.glyphPositions[i]).scale(scale, -scale)
                )
                pen = TransformPen(graphicsPen, transform)
                try:
                    glyph.draw(pen)
                    gc.FillPath(graphicsPen.path, wx.WINDING_RULE)
                except TypeError:
                    log.error("Can't draw glyph %r", glyph)

    def on_KEY_DOWN(self, event: wx.KeyEvent):
        # alt = event.AltDown()
        # cmd = event.CmdDown()
        # ctrl = event.ControlDown()
        # shift = event.ShiftDown()
        unicodeKey = event.GetUnicodeKey()
        if unicodeKey in (wx.WXK_NONE, wx.WXK_BACK, wx.WXK_DELETE, wx.WXK_RETURN):
            key = event.KeyCode
        else:
            key = chr(unicodeKey)
        if key == wx.WXK_HOME and self.caretIndex > 0:
            # move cursor home
            self.caretIndex = 0
        elif key == wx.WXK_LEFT:
            # move cursor one char left
            if self.caretIndex <= 0:
                wx.Bell()
                return
            self.caretIndex -= 1
        elif key == wx.WXK_RIGHT:
            # move cursor one char right
            if self.caretIndex >= self.glyphCount:
                wx.Bell()
                return
            self.caretIndex += 1
        elif key == wx.WXK_END and self.caretIndex < self.glyphCount:
            self.caretIndex = self.glyphCount
        elif key == wx.WXK_BACK:
            if self.caretIndex <= 0:
                wx.Bell()
                return
            self.popGlyph(self.caretIndex - 1)
        elif key == wx.WXK_DELETE:
            if self.caretIndex >= self.glyphCount:
                wx.Bell()
                return
            self.popGlyph(self.caretIndex)
        elif key == wx.WXK_RETURN:
            # insert new line at cursor
            self.insertGlyphAtCaret("\n")
        else:
            event.Skip()

    def on_CHAR(self, event: wx.KeyEvent):
        font = self.currentFont
        if font is None:
            wx.Bell()
            return
        glyphNames = font.unicodeData.get(event.GetUnicodeKey())
        if not glyphNames:
            wx.Bell()
            return
        self.insertGlyphAtCaret(font[glyphNames[0]])

    def on_LEFT_DOWN(self, event: wx.MouseEvent):
        self.setCaretPosition(event.GetX(), event.GetY())

    def on_LEFT_UP(self, event: wx.MouseEvent):
        # print("on_LEFT_UP")
        self.SetFocus()
        event.Skip()

    def on_LEFT_DCLICK(self, event: wx.MouseEvent):
        # print("on_LEFT_DCLICK")
        if not self._glyphs:
            return
        glyph = self._glyphs[self.caretIndex]
        if isinstance(glyph, Glyph):
            glyph.font.showGlyph(glyph.name)
        event.Skip()

    def on_MOUSEWHEEL(self, event: wx.MouseEvent):
        if event.AltDown():
            value = event.GetWheelRotation()
            fontsize = self.fontsize
            if value > 0:
                fontsize += 5
            else:
                fontsize -= 5
            self.fontsize = fontsize
        event.Skip()



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version 3.10.0-4761b0c)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc

###########################################################################
## Class DeleteGlyphsDialogUI
###########################################################################

class DeleteGlyphsDialogUI ( wx.Dialog ):

	def __init__( self, parent ):
		wx.Dialog.__init__ ( self, parent, id = wx.ID_ANY, title = u"Delete Glyphs", pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER )

		self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )

		sizer = wx.BoxSizer( wx.VERTICAL )

		sizer_message = wx.BoxSizer( wx.HORIZONTAL )

		self.bitmap_warn = wx.StaticBitmap( self, wx.ID_ANY, wx.ArtProvider.GetBitmap( wx.ART_WARNING, wx.ART_MESSAGE_BOX ), wx.DefaultPosition, wx.DefaultSize, 0 )
		sizer_message.Add( self.bitmap_warn, 0, wx.ALL, 5 )

		self.lbl_message = wx.StaticText( self, wx.ID_ANY, u"Delete selected glyphs - No Undo!", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_message.SetLabelMarkup( u"Delete selected glyphs - No Undo!" )
		self.lbl_message.Wrap( -1 )

		self.lbl_message.SetFont( wx.Font( wx.NORMAL_FONT.GetPointSize(), wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False, wx.EmptyString ) )

		sizer_message.Add( self.lbl_message, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5 )


		sizer.Add( sizer_message, 1, wx.EXPAND, 5 )


		sizer.Add( ( 0, 5), 0, 0, 0 )

		self.lbl_question = wx.StaticText( self, wx.ID_ANY, u"If deleted Glyphs are used as Component in other Glyphs", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_question.Wrap( -1 )

		sizer.Add( self.lbl_question, 0, wx.ALL, 5 )

		self.radioBtn_decompose = wx.RadioButton( self, wx.ID_ANY, u"Decompose Component", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.radioBtn_decompose.SetValue( True )
		sizer.Add( self.radioBtn_decompose, 0, wx.ALL|wx.EXPAND, 10 )

		self.radioBtn_remove = wx.RadioButton( self, wx.ID_ANY, u"Remove Component from Composite", wx.DefaultPosition, wx.DefaultSize, 0 )
		sizer.Add( self.radioBtn_remove, 0, wx.ALL|wx.EXPAND, 10 )


		sizer.Add( ( 0, 5), 0, 0, 0 )

		buttonSizer = wx.StdDialogButtonSizer()
		self.buttonSizerOK = wx.Button( self, wx.ID_OK )
		buttonSizer.AddButton( self.buttonSizerOK )
		self.buttonSizerCancel = wx.Button( self, wx.ID_CANCEL )
		buttonSizer.AddButton( self.buttonSizerCancel )
		buttonSizer.Realize();

		sizer.Add( buttonSizer, 1, wx.ALL|wx.EXPAND, 5 )


		self.SetSizer( sizer )
		self.Layout()
		sizer.Fit( self )

		self.Centre( wx.BOTH )

	def __del__( self ):
		pass





"""
contour
===============================================================================

Glyph Commands related to contours
"""
from booleanOperations import BooleanOperationManager as Boolean

# from fontToolsTools.pens.cleanupContourPointPen import CleanupContourPointPen

from .base import GlyphCommand
from .parameter import (
    ParamBoolRequired,
    ParamEnumeration,
    ParamFlags,
    ParamFloatRequired,
    ParamIntRequired,
)


class ShiftGlyphCommand(GlyphCommand):
    name = "Shift"
    SHIFT_OUTLINE = 1
    SHIFT_ANCHORS = 2
    SHIFT_COMPONENTS = 4
    parameters = [
        ParamIntRequired("x", "Horizontal shift", 0),
        ParamIntRequired("y", "Vertical shift", 0),
        ParamFlags(
            "f",
            "Apply shift to",
            ["Outline", "Anchros", "Components"],
            SHIFT_OUTLINE | SHIFT_ANCHORS,
        ),
        ParamBoolRequired("p", "Preserve Composites", True),
    ]

    def _execute(self, glyph):
        glyph.moveBy((self.x, self.y))
        if self.p:
            font = glyph.font
            if font and hasattr(font, "componentReferences"):
                componentReferences = font.componentReferences
                if glyph.name in componentReferences:
                    distBack = (-self.x, -self.y)
                    for glyphName in componentReferences[glyph.name]:
                        compositeGlyph = font[glyphName]
                        for component in compositeGlyph.components:
                            if component.baseGlyph == glyph.name:
                                component.moveBy(distBack)


class ScaleGlyphCommand(GlyphCommand):
    name = "Scale"
    SCALE_OUTLINE = 1
    SCALE_ANCHORS = 2
    SCALE_COMPONENTS = 4
    parameters = [
        ParamIntRequired("x", "Horizontal scale [%]", 100),
        ParamIntRequired("y", "Vertical scale [%]", 100),
        ParamEnumeration(
            "c", "Scale center", ["Origin point (0,0)", "Center of glyph"], 0
        ),
        ParamFlags(
            "f",
            "Apply scale to",
            ["Outline", "Anchros", "Components"],
            SCALE_OUTLINE | SCALE_ANCHORS,
        ),
    ]

    def _execute(self, glyph):
        glyph.scaleBy((self.x / 100.0, self.y / 100.0))


class SlantGlyphCommand(GlyphCommand):
    name = "Slant"
    parameters = [
        ParamFloatRequired("a", "Slant angle", 10.0),
        ParamEnumeration(
            "d", "Direction", ["Slant to the right", "Slant to the left"], 0
        ),
        ParamEnumeration(
            "c", "Slant center", ["Origin point (0,0)", "Center of glyph"], 0
        ),
    ]

    def _execute(self, glyph):
        glyph.scaleBy((self.x, self.y))


class ContourDirectionCommand(GlyphCommand):
    name = "Contour direction"
    parameters = [
        ParamEnumeration("d", "Direction", ["PostScript", "TrueType", "Reverse all"], 0)
    ]

    def _execute(self, glyph):
        if self.d == 0:
            glyph.correctContourDirection()
        elif self.d == 1:
            glyph.correctContourDirection(trueType=True)
        elif self.d == 2:
            for contour in glyph:
                contour.reverse()


class RemoveOverlapCommand(GlyphCommand):
    name = "Remove overlap"

    def _execute(self, glyph):
        contours = [c for c in glyph]
        Boolean.union([c for c in glyph], glyph.getPointPen())
        for contour in contours:
            glyph.removeContour(contour)


# class CleanupContourCommand(GlyphCommand):
#     name = "Cleanup Contour"

#     def _execute(self, glyph):
#         pen = CleanupContourPointPen(glyph.getPointPen())
#         contours = list(glyph)
#         glyph.clearContours()
#         for contour in contours:
#             contour.drawPoints(pen)


class RoundCommand(GlyphCommand):
    name = "Round"
    ROUND_OUTLINE = 1
    ROUND_ANCHORS = 2
    ROUND_COMPONENTS = 4
    parameters = [
        ParamFlags(
            "f",
            "Round",
            ["Outline", "Anchros", "Components"],
            ROUND_OUTLINE | ROUND_ANCHORS | ROUND_COMPONENTS,
        )
    ]

    def _execute(self, glyph):
        outline = bool(self.f & self.ROUND_OUTLINE)
        anchors = bool(self.f & self.ROUND_ANCHORS)
        components = bool(self.f & self.ROUND_COMPONENTS)
        glyph.round(None, outline, anchors, components)



"""
toolpan
===============================================================================
"""

import wx

from .base import GlyphTool


class ToolPan(GlyphTool):
    cursor = wx.Cursor(wx.CURSOR_HAND)

    def on_MOTION(self, event):
        GlyphTool.on_MOTION(self, event)
        if event.Dragging():
            canvas = self.canvas
            canvas.Scroll(
                canvas.ViewStart[0] - (self.x - self.xPrev),
                canvas.ViewStart[1] - (self.y - self.yPrev),
            )
            canvas.updateRuler()
        event.Skip()



# -*- coding: utf-8 -*-

###########################################################################
## Python code generated with wxFormBuilder (version 3.10.1-0-g8feb16b3)
## http://www.wxformbuilder.org/
##
## PLEASE DO *NOT* EDIT THIS FILE!
###########################################################################

import wx
import wx.xrc

###########################################################################
## Class FontInfoSuperSubscriptUI
###########################################################################

class FontInfoSuperSubscriptUI ( wx.ScrolledWindow ):

	def __init__( self, parent, id = wx.ID_ANY, pos = wx.DefaultPosition, size = wx.Size( -1,-1 ), style = wx.HSCROLL|wx.VSCROLL, name = wx.EmptyString ):
		wx.ScrolledWindow.__init__ ( self, parent, id = id, pos = pos, size = size, style = style, name = name )

		self.SetExtraStyle( wx.WS_EX_VALIDATE_RECURSIVELY )
		self.SetBackgroundColour( wx.SystemSettings.GetColour( wx.SYS_COLOUR_WINDOW ) )

		sizerMain = wx.BoxSizer( wx.VERTICAL )

		sizerTop = wx.StaticBoxSizer( wx.StaticBox( self, wx.ID_ANY, u"Superscript" ), wx.VERTICAL )

		sizerSuperscript = wx.GridBagSizer( 0, 0 )
		sizerSuperscript.SetFlexibleDirection( wx.HORIZONTAL )
		sizerSuperscript.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

		self.lbl_openTypeOS2SuperscriptXOffset = wx.StaticText( self, wx.ID_ANY, u"x-Offset", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_openTypeOS2SuperscriptXOffset.Wrap( -1 )

		self.lbl_openTypeOS2SuperscriptXOffset.SetMinSize( wx.Size( 170,-1 ) )

		sizerSuperscript.Add( self.lbl_openTypeOS2SuperscriptXOffset, wx.GBPosition( 0, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.txtCtrl_openTypeOS2SuperscriptXOffset = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeOS2SuperscriptXOffset" )
		sizerSuperscript.Add( self.txtCtrl_openTypeOS2SuperscriptXOffset, wx.GBPosition( 0, 1 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.lbl_openTypeOS2SuperscriptXSize = wx.StaticText( self, wx.ID_ANY, u"x-Size", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_openTypeOS2SuperscriptXSize.Wrap( -1 )

		sizerSuperscript.Add( self.lbl_openTypeOS2SuperscriptXSize, wx.GBPosition( 1, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.txtCtrl_openTypeOS2SuperscriptXSize = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeOS2SuperscriptXSize" )
		sizerSuperscript.Add( self.txtCtrl_openTypeOS2SuperscriptXSize, wx.GBPosition( 1, 1 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.lbl_openTypeOS2SuperscriptYOffset = wx.StaticText( self, wx.ID_ANY, u"y-Offset", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_openTypeOS2SuperscriptYOffset.Wrap( -1 )

		sizerSuperscript.Add( self.lbl_openTypeOS2SuperscriptYOffset, wx.GBPosition( 2, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.txtCtrl_openTypeOS2SuperscriptYOffset = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeOS2SuperscriptYOffset" )
		sizerSuperscript.Add( self.txtCtrl_openTypeOS2SuperscriptYOffset, wx.GBPosition( 2, 1 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.lbl_openTypeOS2SuperscriptYSize = wx.StaticText( self, wx.ID_ANY, u"y-Size", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_openTypeOS2SuperscriptYSize.Wrap( -1 )

		sizerSuperscript.Add( self.lbl_openTypeOS2SuperscriptYSize, wx.GBPosition( 3, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.txtCtrl_openTypeOS2SuperscriptYSize = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeOS2SuperscriptYSize" )
		sizerSuperscript.Add( self.txtCtrl_openTypeOS2SuperscriptYSize, wx.GBPosition( 3, 1 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )


		sizerSuperscript.AddGrowableCol( 1 )

		sizerTop.Add( sizerSuperscript, 1, wx.EXPAND, 5 )


		sizerMain.Add( sizerTop, 0, wx.ALL|wx.EXPAND, 5 )

		sizerBottom = wx.StaticBoxSizer( wx.StaticBox( self, wx.ID_ANY, u"Subscript" ), wx.VERTICAL )

		sizerSubscript = wx.GridBagSizer( 0, 0 )
		sizerSubscript.SetFlexibleDirection( wx.HORIZONTAL )
		sizerSubscript.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

		self.lbl_openTypeOS2SubscriptXOffset = wx.StaticText( self, wx.ID_ANY, u"x-Offset", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_openTypeOS2SubscriptXOffset.Wrap( -1 )

		self.lbl_openTypeOS2SubscriptXOffset.SetMinSize( wx.Size( 170,-1 ) )

		sizerSubscript.Add( self.lbl_openTypeOS2SubscriptXOffset, wx.GBPosition( 0, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.txtCtrl_openTypeOS2SubscriptXOffset = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeOS2SubscriptXOffset" )
		sizerSubscript.Add( self.txtCtrl_openTypeOS2SubscriptXOffset, wx.GBPosition( 0, 1 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.lbl_openTypeOS2SubscriptXSize = wx.StaticText( self, wx.ID_ANY, u"x-Size", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_openTypeOS2SubscriptXSize.Wrap( -1 )

		sizerSubscript.Add( self.lbl_openTypeOS2SubscriptXSize, wx.GBPosition( 1, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.txtCtrl_openTypeOS2SubscriptXSize = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeOS2SubscriptXSize" )
		sizerSubscript.Add( self.txtCtrl_openTypeOS2SubscriptXSize, wx.GBPosition( 1, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )

		self.lbl_openTypeOS2SubscriptYOffset = wx.StaticText( self, wx.ID_ANY, u"y-Offset", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_openTypeOS2SubscriptYOffset.Wrap( -1 )

		sizerSubscript.Add( self.lbl_openTypeOS2SubscriptYOffset, wx.GBPosition( 2, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.txtCtrl_openTypeOS2SubscriptYOffset = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeOS2SubscriptYOffset" )
		sizerSubscript.Add( self.txtCtrl_openTypeOS2SubscriptYOffset, wx.GBPosition( 2, 1 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.lbl_openTypeOS2SubscriptYSize = wx.StaticText( self, wx.ID_ANY, u"y-Size", wx.DefaultPosition, wx.DefaultSize, 0 )
		self.lbl_openTypeOS2SubscriptYSize.Wrap( -1 )

		sizerSubscript.Add( self.lbl_openTypeOS2SubscriptYSize, wx.GBPosition( 3, 0 ), wx.GBSpan( 1, 1 ), wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5 )

		self.txtCtrl_openTypeOS2SubscriptYSize = wx.TextCtrl( self, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0, wx.DefaultValidator, u"openTypeOS2SubscriptYSize" )
		sizerSubscript.Add( self.txtCtrl_openTypeOS2SubscriptYSize, wx.GBPosition( 3, 1 ), wx.GBSpan( 1, 1 ), wx.ALL, 5 )


		sizerSubscript.AddGrowableCol( 1 )

		sizerBottom.Add( sizerSubscript, 1, wx.EXPAND, 5 )


		sizerMain.Add( sizerBottom, 0, wx.ALL|wx.EXPAND, 5 )


		self.SetSizer( sizerMain )
		self.Layout()
		sizerMain.Fit( self )

	def __del__( self ):
		pass





"""
newGlyphDialog
===============================================================================
"""
import re
import string

import wx

from .newGlyphDialogUI import NewGlyphDialogUI


class GlyphNameValidator(wx.Validator):
    validChars = string.ascii_letters + string.digits + "_."
    validPattern = r"^[a-zA-Z_][a-zA-Z0-9_.]*$"

    def __init__(self):
        super().__init__()
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def Clone(self):
        return GlyphNameValidator()

    def Validate(self, win):
        textCtrl = self.GetWindow()
        text = textCtrl.GetValue()
        if len(text) == 0:
            wx.MessageBox("Glyph name must not be empty!", "Error")
            textCtrl.SetBackgroundColour("pink")
            textCtrl.SetFocus()
            textCtrl.Refresh()
            return False
        else:
            textCtrl.SetBackgroundColour(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW)
            )
            textCtrl.Refresh()
            return True

    def TransferToWindow(self):
        return True

    def TransferFromWindow(self):
        return True

    def OnChar(self, event):
        key = event.GetKeyCode()
        if (
            key < wx.WXK_SPACE
            or key == wx.WXK_DELETE
            or key > 255
            or chr(key) in self.validChars
        ):
            event.Skip()
            return
        if not wx.Validator.IsSilent():
            wx.Bell()
        return


class NewGlyphDialog(NewGlyphDialogUI):
    def __init__(self):
        super().__init__(wx.GetApp().panelManager.documentNotebook)
        self.textCtrl_name.SetValidator(GlyphNameValidator())

    @property
    def glyphname(self):
        if self.textCtrl_name.Value:
            return self.textCtrl_name.Value

    @property
    def unicodes(self):
        if self.textCtrl_unicode.Value:
            return [int(u, 16) for u in self.textCtrl_unicode.Value.split()]
        return []



"""
deleteGlyphsDialog
===============================================================================
"""
import wx

from .deleteGlyphsDialogUI import DeleteGlyphsDialogUI


class DeleteGlyphsDialog(DeleteGlyphsDialogUI):
    def __init__(self, glyphCount):
        super().__init__(parent=wx.GetApp().TopWindow)
        self.lbl_message.SetLabelMarkup(
            f"Delete {glyphCount} selected glyphs - No Undo!"
        )

    @property
    def component_handling(self):
        if self.radioBtn_decompose.Value:
            return "decompose"
        return "remove"



"""
anchor
===============================================================================

Glyph Commands related to glyph anchors
"""
import wx

from wbDefcon import Anchor, Color, Glyph

from .base import GlyphCommand
from .parameter import (
    ParamBoolRequired,
    ParamStrRequired,
    ParamColour,
    ParamIntRequired,
)


class AddAnchorCommand(GlyphCommand):
    name = "Add anchor"
    parameters = [
        ParamIntRequired("x", "Horizontal position (x)", 0),
        ParamIntRequired("y", "Vertical position (y)", 0),
        ParamStrRequired("n", "Name"),
        ParamColour("c", "Colour", wx.WHITE),
        ParamBoolRequired("r", "Replace existing anchor (by name)", False),
    ]

    def _execute(self, glyph: Glyph):
        for anchor in glyph.anchors:
            if anchor.name == self.n:
                if self.r:
                    anchor.holdNotifications()
                    anchor.x = self.x
                    anchor.y = self.y
                    if self.c != wx.WHITE:
                        anchor.color = Color.from_wx(self.c)
                    anchor.releaseHeldNotifications()
                return
        anchor = Anchor(anchorDict=dict(name=self.n, x=self.x, y=self.y))
        if self.c != wx.WHITE:
            anchor.color = Color.from_wx(self.c)
        glyph.appendAnchor(anchor)


class RemoveAnchorCommand(GlyphCommand):
    name = "Remove anchor (by name)"
    parameters = [ParamStrRequired("n", "Name")]

    def _execute(self, glyph: Glyph):
        for anchor in reversed(glyph.anchors):
            if anchor.name == self.n:
                glyph.removeAnchor(anchor)


class ClearAnchorCommand(GlyphCommand):
    name = "Clear all anchors"

    def _execute(self, glyph: Glyph):
        glyph.clearAnchors()



"""
parameter
===============================================================================

Glyph-Command-Parameters
"""
import logging

import wx
from wx.propgrid import (
    BoolProperty,
    ColourProperty,
    EditEnumProperty,
    EnumProperty,
    FileProperty,
    FlagsProperty,
    FloatProperty,
    IntProperty,
    StringProperty,
)


log = logging.getLogger(__name__)


class GlyphCmdParameter:
    """Base class for Glyph-Command-Parameters"""

    def __init__(self, name, type, label=None, default=None, allowNone=True):
        self.name = name
        self.type = type
        if label:
            self.label = label
        else:
            self.label = self.name
        self.allowNone = allowNone
        if default is not None:
            assert isinstance(default, self.type)
        elif not self.allowNone:
            assert default is not None
        self.default = default

    def __repr__(self):
        return "<%s: name=%s>" % (self.__class__.__name__, self.name)

    def fromString(self, value):
        log.debug("GlyphCmdParameter.fromString(%r)", value)
        return self.type(value)

    def toString(self, value):
        log.debug("GlyphCmdParameter %r toString(%r)", self.name, value)
        return str(value)

    def pgProperty(self, value):
        return NotImplemented

class ParamIntRequired(GlyphCmdParameter):
    """Required integer parameter"""

    def __init__(self, name, label=None, default=0):
        super().__init__(name, int, label, default, allowNone=False)

    def pgProperty(self, value):
        prop = IntProperty(self.label, self.name, value)
        prop.SetEditor("SpinCtrl")
        return prop


class ParamFloatRequired(GlyphCmdParameter):
    """Required float parameter"""

    def __init__(self, name, label=None, default=0):
        super().__init__(name, float, label, default, allowNone=False)

    def pgProperty(self, value):
        return FloatProperty(self.label, self.name, value)


class ParamStr(GlyphCmdParameter):
    """Optional string parameter"""

    def __init__(self, name, label=None, default=None):
        super().__init__(name, str, label, default, allowNone=True)

    def pgProperty(self, value):
        return StringProperty(self.label, self.name, value)


class ParamStrRequired(GlyphCmdParameter):
    """Required string parameter"""

    def __init__(self, name, label=None, default=""):
        super().__init__(name, str, label, default, allowNone=False)

    def pgProperty(self, value):
        return StringProperty(self.label, self.name, value)


class ParamBoolRequired(GlyphCmdParameter):
    """Required boolean parameter"""

    def __init__(self, name, label=None, default=False):
        super().__init__(name, bool, label, default, allowNone=False)

    def pgProperty(self, value):
        prop = BoolProperty(self.label, self.name, value)
        prop.SetEditor("CheckBox")
        return prop


class ParamEnumeration(GlyphCmdParameter):
    def __init__(self, name, label=None, choices=None, default=0):
        super().__init__(name, int, label, default, allowNone=False)
        self.labels = choices or []

    def pgProperty(self, value):
        return EnumProperty(
            self.label, self.name, self.labels, range(len(self.labels)), value
        )


class ParamEditableEnumeration(GlyphCmdParameter):
    def __init__(self, name, label=None, choices=None, default=None):
        self.labels = choices or []
        if isinstance(default, int) and default in range(len(self.labels)):
            default = self.labels[default]
        super().__init__(name, str, label=label, default=default, allowNone=False)

    def toString(self, value):
        if isinstance(value, str):
            return value
        if isinstance(value, int) and value in range(len(self.labels)):
            return self.labels[value]

    def pgProperty(self, value):
        if isinstance(value, int) and value in range(len(self.labels)):
            value = self.labels[value]
        return EditEnumProperty(
            self.label, self.name, self.labels, range(len(self.labels)), value
        )


class ParamFlags(GlyphCmdParameter):
    def __init__(self, name, label=None, choices=None, default=0):
        super().__init__(name, int, label, default, allowNone=False)
        self.labels = choices or []

    def pgProperty(self, value):
        return FlagsProperty(
            self.label,
            self.name,
            self.labels,
            [2 ** i for i in range(len(self.labels))],
            value,
        )


class ParamColour(GlyphCmdParameter):
    def __init__(self, name, label=None, default=wx.RED):
        super().__init__(name, wx.Colour, label, default, allowNone=False)

    def fromString(self, value):
        colour = self.type()
        colour.Set(value)
        return colour

    def toString(self, value):
        return value.GetAsString(wx.C2S_CSS_SYNTAX)

    def pgProperty(self, value):
        return ColourProperty(self.label, self.name, value)


class ParamFilepathRequired(ParamStrRequired):
    def pgProperty(self, value):
        return FileProperty(self.label, self.name, value)



"""
findGlyphDialog
===============================================================================
"""
import logging

import wx

from wbDefcon import Font, Layer
from wbFontParts import RFont, RLayer

from .findGlyphDialogUI import FindGlyphDialogUI

log = logging.getLogger(__name__)

###########################################################################
## Class FindGlyphDialog
###########################################################################


class FindGlyphDialog(FindGlyphDialogUI):
    """
    Dialog to search for a glyph in a font or layer.
    """
    def __init__(self, parent, font_or_layer, allowSelection=False):
        super().__init__(parent)
        self.sizerDialogButtonsOK.SetDefault()
        self._layer = None
        self.layer = font_or_layer
        self.glyphnames = self.layer.keys()
        self.characters = {"%04X" % k: v[0] for k, v in self.layer.unicodeData.items()}
        self.dataFind = []
        self.selctedGlyph = None
        self.allowSelection = allowSelection
        self.textCtrl_findValue.SetFocus()

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, value):
        if isinstance(value, Font):
            self._layer = value.layers.defaultLayer
        elif isinstance(value, RFont):
            self._layer = value.naked().layers.defaultLayer
        elif isinstance(value, Layer):
            self._layer = value
        elif isinstance(value, RLayer):
            self._layer = value.naked()
        else:
            log.error("got font_or_layer as %r", value)

    def updateFindData(self):
        self.selctedGlyph = None
        self.listCtrl_glyphs.SetItemCount(0)
        text = self.textCtrl_findValue.GetValue()
        if text:
            findAttr = self.choice_findAttr.StringSelection
            findCompare = self.choice_findCompare.StringSelection
            if findAttr == "Name":
                if findCompare == "equals to":
                    self.dataFind = sorted([n for n in self.glyphnames if n == text])
                elif findCompare == "starts with":
                    self.dataFind = sorted(
                        [
                            (n, self.characters.get(n, ""))
                            for n in self.glyphnames
                            if n.startswith(text)
                        ]
                    )
                elif findCompare == "ends with":
                    self.dataFind = sorted(
                        [
                            (n, self.characters.get(n, ""))
                            for n in self.glyphnames
                            if n.endswith(text)
                        ]
                    )
                elif findCompare == "contains":
                    self.dataFind = sorted(
                        [
                            (n, self.characters.get(n, ""))
                            for n in self.glyphnames
                            if text in n
                        ]
                    )
                else:
                    self.dataFind = []
            elif findAttr == "Unicode":
                sortByUni = lambda i: i[1]
                if findCompare == "equals to":
                    self.dataFind = sorted(
                        [(v, k) for k, v in self.characters.items() if k == text],
                        key=sortByUni,
                    )
                elif findCompare == "starts with":
                    self.dataFind = sorted(
                        [
                            (v, k)
                            for k, v in self.characters.items()
                            if k.startswith(text)
                        ],
                        key=sortByUni,
                    )
                elif findCompare == "ends with":
                    self.dataFind = sorted(
                        [
                            (v, k)
                            for k, v in self.characters.items()
                            if k.endswith(text)
                        ],
                        key=sortByUni,
                    )
                elif findCompare == "contains":
                    self.dataFind = sorted(
                        [(v, k) for k, v in self.characters.items() if text in k],
                        key=sortByUni,
                    )
                else:
                    self.dataFind = []
        else:
            self.dataFind = []
        self.listCtrl_glyphs.SetItemCount(len(self.dataFind))
        self.listCtrl_glyphs.SetItemState(
            0, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED
        )

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_TEXT_ENTER(self, event):
        if self.selctedGlyph is not None:
            self.EndModal(wx.ID_OK)

    def onfindAttr_CHOICE(self, event):
        self.updateFindData()

    def onfindCompare_CHOICE(self, event):
        self.updateFindData()

    def on_TEXT(self, event):
        self.updateFindData()

    def on_LIST_ITEM_SELECTED(self, event):
        self.selctedGlyph = self.listCtrl_glyphs.GetItemText(event.Index)
        self.bitmapGlyph.SetBitmap(
            self.layer[self.selctedGlyph].getRepresentation(
                "bitmap", width=72, height=72, captionHeight=0
            )
        )
        event.Skip()

    def on_LIST_ITEM_ACTIVATED(self, event):
        self.selctedGlyph = self.listCtrl_glyphs.GetItemText(event.Index)
        self.bitmapGlyph.SetBitmap(
            self.layer[self.selctedGlyph].getRepresentation(
                "bitmap", width=72, height=72, captionHeight=0
            )
        )
        if self.selctedGlyph is not None:
            self.EndModal(wx.ID_OK)

    def on_buttonSelect(self, event):
        self.layer.selectedGlyphNames = [i[0] for i in self.dataFind]

    def update_buttonSelect(self, event):
        event.Enable(self.allowSelection and len(self.dataFind) > 0)



"""
toolknife
===============================================================================
"""
import logging
import math

import wx
from booleanOperations import BooleanOperationManager as Boolean
from wbBase.tools import get_wxBrush, get_wxPen
from wbDefcon import Contour, Point
from wbDefcon.pens.lineIntersectionPen import LineIntersectionPen

from ..cursor import cursors
from .base import WORKING, GlyphTool

log = logging.getLogger(__name__)


class LineContourIntersectionPen(LineIntersectionPen):
    """
    Works like LineIntersectionPen but ignores components.
    """

    def addComponent(self, glyphName, transformation):
        pass


class ToolKnife(GlyphTool):
    cursor = cursors["Knife"]
    pen = get_wxPen(wx.Colour(128, 128, 150, 128))
    brush = get_wxBrush(wx.Colour(128, 128, 150, 64))
    cutpen = get_wxPen(wx.Colour(255, 64, 64, 255))

    def __init__(self, parent):
        super().__init__(parent)
        self.intersections = []
        self.lineIntersectionPen = LineContourIntersectionPen()
        self.knifeStartPoint = None
        self.knifeEndPoint = None
        self.cutStartPoint = None
        self.cutEndPoint = None

    def reset(self):
        super().reset()
        self.knifeStartPoint = None
        self.knifeEndPoint = None
        self.cutStartPoint = None
        self.cutEndPoint = None

    @property
    def knifeLine(self):
        if self.knifeStartPoint and self.knifeEndPoint:
            return (self.knifeStartPoint, self.knifeEndPoint)
        return None

    @staticmethod
    def midpoint(p1, p2):
        return ((p1[0] + p2[0]) / 2, (p1[1] + p2[1]) / 2)

    # -------------------------------------------------------------------------
    # Event handler methods
    # -------------------------------------------------------------------------

    def on_LEFT_DOWN(self, event):
        super().on_LEFT_DOWN(event)
        glyph = self.glyph
        if glyph is not None:
            self.lineIntersectionPen.glyphSet = glyph.layer
            self.knifeStartPoint = (
                round(self.canvas.screenToCanvasX(self.x)),
                round(self.canvas.screenToCanvasY(self.y)),
            )
            self.state = WORKING
        event.Skip()

    def on_LEFT_UP(self, event):
        glyph = self.glyph
        if glyph is not None:
            if self.intersections and self.cutStartPoint and self.cutEndPoint:
                p0 = self.cutStartPoint
                p1 = self.cutEndPoint
                distance = self.distance(p0, p1)
                cutRect = Contour()
                d = 0.01
                cutRect.appendPoint(Point((p0[0], p0[1] - d), "line"))
                cutRect.appendPoint(
                    Point(
                        (p0[0] + distance, p0[1] - d),
                        "line",
                    )
                )
                cutRect.appendPoint(
                    Point(
                        (p0[0] + distance, p0[1] + d),
                        "line",
                    )
                )
                cutRect.appendPoint(Point((p0[0], p0[1] + d), "line"))
                dx = p1[0] - p0[0]
                dy = p1[1] - p0[1]
                angle = math.degrees(math.atan2(dy, dx))
                cutRect.rotateBy(angle, p0)
                contours = list(glyph)
                glyph.disableNotifications()
                glyph.undoManager.saveState()
                glyph.clearContours()
                Boolean.difference(
                    contours,
                    [cutRect],
                    glyph.getPointPen(),
                )
                glyph.enableNotifications()
                glyph.postNotification("Glyph.ContoursChanged")
        super().on_LEFT_UP(event)

    def on_MOTION(self, event):
        super().on_MOTION(event)
        if event.Dragging() and self.state == WORKING:
            x = self.x
            y = self.y
            if self.shiftDown:
                dx = abs(self.xClick - self.x)
                dy = abs(self.yClick - self.y)
                if dx > dy:
                    y = self.yClick
                else:
                    x = self.xClick
            self.knifeEndPoint = (
                round(self.canvas.screenToCanvasX(x)),
                round(self.canvas.screenToCanvasY(y)),
            )
            points = []
            glyph = self.glyph
            if glyph is not None:
                line = self.knifeLine
                if line:
                    pen = self.lineIntersectionPen
                    if line != pen.line:
                        pen.line = line
                        glyph.draw(pen)
                points = list(pen.intersections) + [
                    self.knifeStartPoint,
                    self.knifeEndPoint,
                ]
                points.sort()
                if glyph.pointInside(points[0]):
                    try:
                        self.cutStartPoint = self.midpoint(points[1], points[2])
                    except IndexError:
                        self.cutStartPoint = None
                    points = points[2:]
                else:
                    self.cutStartPoint = points[0]
                    points = points[1:]
                if points:
                    if glyph.pointInside(points[-1]):
                        try:
                            self.cutEndPoint = self.midpoint(points[-3], points[-2])
                        except IndexError:
                            self.cutEndPoint = None
                        points = points[:-2]
                    else:
                        self.cutEndPoint = points[-1]
                        points = points[:-1]
                else:
                    self.cutEndPoint = None
            self.intersections = points
            self.canvas.Refresh()
        event.Skip()

    def draw(self, gc):
        if self.state == WORKING and self.knifeStartPoint and self.knifeEndPoint:
            transformPoint = self.canvas.transform.transformPoint
            start = transformPoint(self.knifeStartPoint)
            end = transformPoint(self.knifeEndPoint)
            gc.SetPen(self.pen)
            gc.StrokeLine(start[0], start[1], end[0], end[1])
            if self.intersections:
                gc.SetPen(self.cutpen)
                for i in range(0, len(self.intersections), 2):
                    try:
                        p0 = transformPoint(self.intersections[i])
                        p1 = transformPoint(self.intersections[i + 1])
                        gc.StrokeLine(p0[0], p0[1], p1[0], p1[1])
                    except IndexError:
                        continue
                s = 2
                gc.SetPen(self.pen)
                gc.SetBrush(self.brush)
                for point in self.intersections:
                    try:
                        x, y = transformPoint(point)
                        x = round(x)
                        y = round(y)
                    except ValueError:
                        continue
                    gc.DrawEllipse(x - s, y - s, 2 * s, 2 * s)



from __future__ import annotations

import logging
import os
import weakref
from typing import TYPE_CHECKING, Type, Optional

import wx
from wbBase.document import dbg
from wbBase.document.notebook import DocumentPageMixin
from wbBase.document.view import View
from wbDefcon import Font, Glyph

from ..control.glyphEditor.editPanel import GlyphEditPanel

if TYPE_CHECKING:
    from ..control.glyphEditor.canvas import Canvas
    from ..document import UfoDocument

log = logging.getLogger(__name__)


class UfoGlyphWindow(wx.Notebook, DocumentPageMixin):
    # canvas: Canvas
    document:UfoDocument

    def __init__(self, parent, doc:UfoDocument, view:UfoGlyphView):
        wx.Notebook.__init__(
            self,
            parent,
            id=wx.ID_ANY,
            pos=wx.DefaultPosition,
            size=wx.DefaultSize,
            style=wx.NB_FIXEDWIDTH | wx.NO_BORDER,
            name="UfoGlyphWindow",
        )
        DocumentPageMixin.__init__(self, doc, view)
        self._glyph = None
        self._font = None
        # canvas page
        self.editPanel = GlyphEditPanel(self, doc=doc)
        # self.canvas = self.editPanel.canvas  # is this realy needed?
        self.AddPage(self.editPanel, "Canvas", False)

        self.Layout()
        self.font = self.document.font

    @property
    def canvas(self) -> Canvas:
        return self.editPanel.canvas

    @property
    def font(self) -> Optional[Font]:
        if self._font is None:
            return None
        return self._font()

    @font.setter
    def font(self, font: Font):
        assert isinstance(font, Font)
        assert self._font is None
        self.canvas.font = font
        self._font = weakref.ref(font)

    @font.deleter
    def font(self):
        del self.canvas.font
        self._font = None

    @property
    def glyph(self) -> Optional[Glyph]:
        return self._glyph

    @glyph.setter
    def glyph(self, glyph: Glyph):
        dbg(f"UfoGlyphWindow set glyph {glyph}")
        if glyph is None:
            del self.glyph
            return
        assert isinstance(glyph, Glyph)
        assert glyph.font == self.font
        if self.glyph != glyph:
            del self.glyph
            glyph.addObserver(self, "handleNotification", "Glyph.NameChanged")
            self.canvas.glyph = glyph
            self._glyph = glyph
            # self.title = "Glyph - %s" % glyph.name
            if self.font.path:
                self.title = (
                    f"Glyph - {glyph.name} of {os.path.basename(self.font.path)}"
                )
            else:
                self.title = f"Glyph - {glyph.name}"

    @glyph.deleter
    def glyph(self):
        if isinstance(self._glyph, Glyph):
            self._glyph.removeObserver(self, "Glyph.NameChanged")
        del self.canvas.glyph
        self._glyph = None

    def handleNotification(self, notification):
        if notification.name == "Glyph.NameChanged":
            if self.font.path:
                self.title = f"Glyph - {notification.data['newValue']} of {os.path.basename(self.font.path)}"
            else:
                self.title = f"Glyph - {notification.data['newValue']}"

    def Destroy(self):
        # self.canvas = None
        self.editPanel.Destroy()
        return super().Destroy()


class UfoGlyphView(View):
    frameType = UfoGlyphWindow
    _frame: UfoGlyphWindow
    frame: UfoGlyphWindow

    def __init__(self):
        super().__init__()
        self._typeName = "UFO Glyph View"

    @property
    def font(self) -> Optional[Font]:
        if self._document and self._document._data:
            return self._document._data
        return None

    @property
    def glyph(self) -> Optional[Glyph]:
        if self._frame:
            return self._frame.glyph
        return None

    @glyph.setter
    def glyph(self, glyph:Glyph):
        dbg(f"UfoGlyphView set glyph {glyph}")
        if glyph is None:
            del self.glyph
            return
        assert isinstance(glyph, Glyph)
        assert glyph.font == self.font
        if self.glyph != glyph:
            del self.glyph
            self._frame.glyph = glyph

    @glyph.deleter
    def glyph(self):
        del self._frame.glyph

    def OnCreate(self, doc:UfoDocument, flags:int):
        dbg(f'UfoGlyphView.OnCreate(doc="{doc}", flags={flags})', indent=1)
        self._document = doc
        self._frame = self.frameType(self.documentNotebook, doc, self)
        self.documentNotebook.AddPage(
            self._frame,
            doc.printableName,
            True,
            wx.ArtProvider.GetBitmap("G", wx.ART_FRAME_ICON),
        )
        dbg(indent=0)
        return True

    def Destroy(self):
        dbg("UfoGlyphView.Destroy", indent=1)
        super().Destroy()
        dbg(indent=0)
